#	Copyright (C) 2017 Battelle Memorial Institute
# file: process_houses.py; focus on HVAC
import json;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;

casefiles = [['SGIP1a','red'],
						 ['SGIP1b','blue']]
#						 ['SGIP1c','green'],
#						 ['SGIP1d','magenta'],
#						 ['SGIP1e','cyan'],
#						 ['SGIP1ex','orange']]

def MakePlotData(root):
	lp = open (root + "_agent_dict.json")
	dict = json.loads(lp.read())
	lp.close()

	a_keys = list(dict['markets'].keys())
	a_keys.sort()
	lp_a = open ('auction_' + root + '_metrics.json')
	lst_a = json.loads(lp_a.read())
	lp_a.close()
	lst_a.pop('StartTime')
	meta_a = lst_a.pop('Metadata')

	times = list(map(int,list(lst_a.keys())))
	times.sort()
	hrs = np.array(times, dtype=np.float)
	denom = 3600.0
	hrs /= denom

	time_key = str(times[0])

	for key, val in meta_a.items():
		if key == 'clearing_price':
			CLEAR_IDX = val['index']
			CLEAR_UNITS = val['units']
		elif key == 'clearing_type':
			TYPE_IDX = val['index']


	data_a = np.empty(shape=(len(a_keys), len(times), len(lst_a[time_key][a_keys[0]])), dtype=np.float)
	j = 0
	for key in a_keys:
		i = 0
		for t in times:
			ary = lst_a[str(t)][a_keys[j]]
			data_a[j, i,:] = ary
			i = i + 1
		j = j + 1

	cprice = data_a[0,:,CLEAR_IDX]
	ctype = data_a[0,:,TYPE_IDX]
	return hrs, cprice, ctype

def compare_auction ():
	# display a plot
	tmin = 0.0
	tmax = 48.0
	xticks = [0,6,12,18,24,30,36,42,48]

	fig, ax = plt.subplots(2, 1, sharex = 'col')

	for root in casefiles:
		print ('Processing', root[0])
		hrs, cprice, ctype = MakePlotData(root[0])
		ax[0].plot(hrs, cprice, color=root[1], label=root[0])
		ax[1].plot(hrs, ctype, color=root[1], label=root[0])

	ax[0].set_title ('Cleared Price')
	ax[0].set_ylabel ('$')

	ax[1].set_title ('Cleared Type')

	ax[1].set_xlabel ('Hours')

	ax[0].grid()
	ax[0].legend()
	ax[0].set_xlim(tmin,tmax)
	ax[0].set_xticks(xticks)

	ax[1].grid()
	ax[1].legend()
	ax[1].set_xlim(tmin,tmax)
	ax[1].set_xticks(xticks)

	plt.show()





#import sys
#from datetime import datetime
#from datetime import timedelta
#StartTime = '2013-07-01 00:00:00 PST+8PDT'
#time_fmt = '%Y-%m-%d %H:%M:%S %Z'
#StartTime = '2013-07-01 00:00:00 -0800'
#time_fmt = '%Y-%m-%d %H:%M:%S %z'
#dt_now = datetime.strptime (StartTime, time_fmt)

# should not begin with a number, or contain '-' for FNCS
#def gld_strict_name(val):
#  if val[0].isdigit():
#    val = 'gld_' + val
#  return val.replace ('-', '_')

#oldname = '9R1-12-47-1_tn_1_hse_1'
#newname = gld_strict_name (oldname)
#print (oldname, newname)

# step 1
import tesp_support.tesp_config as tesp
tesp.show_tesp_config()

# step 2
#import tesp_support.api as tesp
#tesp.make_tesp_case ('Demo.json')

# step 3
#import tesp_support.api as tesp
#tesp.make_monte_carlo_cases ('Demo.json')

# step 4
#import tesp_support.tesp_monitor as tesp
#tesp.show_tesp_monitor()




# Copyright (C) 2018-2019 Battelle Memorial Institute
# file: tesp_monitor_ercot.py
"""Presents a GUI to launch a TESP simulation and monitor its progress

This version differs from the one in *tesp_monitor*, in that the
user can select the FNCS federate and topic to plot. The number of
monitored plots is still fixed at 4.

Public Functions:
  :show_tesp_monitor: Initializes and runs the monitor GUI

References:
  `Graphical User Interfaces with Tk <https://docs.python.org/3/library/tk.html>`_

  `Matplotlib Animation <https://matplotlib.org/api/animation_api.html>`_
"""
import sys
import json
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog
from tkinter import messagebox
import subprocess
import os
try:
  import tesp_support.fncs as fncs
except:
  pass
import tesp_support.helpers as helpers
import time

import numpy as np;
import matplotlib;

matplotlib.use('TkAgg');
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk;
from matplotlib.figure import Figure;
from matplotlib.lines import Line2D
import matplotlib.animation as animation
import matplotlib.pyplot as plt;
import yaml


class TespMonitorGUI:
    """Version of the monitor GUI that hosts 4 choosable plots

    Args:
        master:

    Attributes:
        root (Tk): the TCL Tk toolkit instance
        top (Window): the top-level TCL Tk Window
        labelvar (StringVar): used to display the monitor JSON configuration file path
        plot0 (ChoosablePlot): first plot
        plot1 (ChoosablePlot): second plot
        plot2 (ChoosablePlot): third plot
        plot3 (ChoosablePlot): fourth plot
        topicDict:
        plots (Frame):
        scrollbar (Scrollbar):
        frameInCanvas (Frame):
        canvas (FigureCanvasTkAgg): a TCL Tk canvas that can host Matplotlib
        bFNCSactive (Boolean): True if a TESP simulation is running with other FNCS federates, False if not
    """
    def __init__(self, master):
        self.root = master
        self.root.protocol('WM_DELETE_WINDOW', self.on_closing)
        self.topicDict = {}
        # self.topicDict = {'pypower/LMP_Bus8' : 'LMP8',
        #                   'gridlabdBus8/distribution_load' : 'distribution_load8',
        #                   'pypower/three_phase_voltage_Bus1' : 'vpos1',
        #                   'pypower/three_phase_voltage_Bus8' : 'vpos8'}

        self.top = self.root.winfo_toplevel()
        self.top.rowconfigure(0, weight=1)
        self.top.columnconfigure(0, weight=1)

        self.root.rowconfigure(0, weight=0)
        self.root.rowconfigure(1, weight=1)
        self.root.columnconfigure(0, weight=0)
        self.root.columnconfigure(1, weight=0)
        self.root.columnconfigure(2, weight=0)
        self.root.columnconfigure(3, weight=0)
        self.root.columnconfigure(4, weight=1)

        ttk.Style().configure('TButton', foreground='blue')

        btn = ttk.Button(self.root, text='Open...', command=self.OpenConfig)
        btn.grid(row=0, column=0, sticky=tk.NSEW)
        btn = ttk.Button(self.root, text='Start All', command=self.launch_all)
        btn.grid(row=0, column=1, sticky=tk.NSEW)
        btn = ttk.Button(self.root, text='Kill All', command=self.kill_all)
        btn.grid(row=0, column=2, sticky=tk.NSEW)
        btn = ttk.Button(self.root, text='Quit', command=self.Quit)
        btn.grid(row=0, column=3, sticky=tk.NSEW)
        self.labelvar = tk.StringVar()
        self.labelvar.set('Case')
        lab = ttk.Label(self.root, textvariable=self.labelvar, relief=tk.RIDGE)
        lab.grid(row=0, column=4, sticky=tk.NSEW)

        self.plots = tk.Frame(self.root, bg="#ffffff")
        self.plots.grid(row=1, column=0, columnspan=5, sticky=tk.NSEW)
        self.plots.grid_rowconfigure(0, weight=1)
        self.plots.grid_columnconfigure(0, weight=1)

        self.canvas = tk.Canvas(self.plots, borderwidth=0, background="#ffffff")
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        self.scrollbar = tk.Scrollbar(self.plots, command=self.canvas.yview)
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)

        self.canvas.configure(yscrollcommand=self.scrollbar.set)

        self.frameInCanvas = tk.Frame(self.canvas, background="#ffffff")
        self.frameInCanvas.grid(row=0, column=0, sticky='news')
        self.canvas.create_window((0, 0), window=self.frameInCanvas, anchor=tk.NW, tags="self.frameInCanvas")
        self.frameInCanvas.bind("<Configure>", self.onFrameConfigure)

        self.canvas.rowconfigure(0, weight=0)
        self.canvas.columnconfigure(0, weight=0)
        self.frameInCanvas.rowconfigure(0, weight=1)
        self.frameInCanvas.rowconfigure(1, weight=1)
        self.frameInCanvas.rowconfigure(2, weight=1)
        self.frameInCanvas.rowconfigure(3, weight=2)
        self.frameInCanvas.columnconfigure(0, weight=1)

        self.plot0 = ChoosablePlot(self.frameInCanvas, color='green', topicDict=self.topicDict)
        self.plot0.grid(row=0, sticky=tk.NSEW)
        self.plot1 = ChoosablePlot(self.frameInCanvas, color='red', topicDict=self.topicDict)
        self.plot1.grid(row=1, sticky=tk.NSEW)
        self.plot2 = ChoosablePlot(self.frameInCanvas, color='blue', topicDict=self.topicDict)
        self.plot2.grid(row=2, sticky=tk.NSEW)
        self.plot3 = ChoosablePlot(self.frameInCanvas, color='magenta', xLabel='Hours', topicDict=self.topicDict)
        self.plot3.grid(row=3, sticky=tk.NSEW, pady=(0, 50))
        # self.plot0.pack()
        # self.plot1.pack()
        # self.plot2.pack()
        # self.plot3.pack()
        # self.fig0, self.ax0 = plt.subplots(1,1)
        # cax = plt.gca()
        # cax.axes.xaxis.set_ticklabels([])
        # self.fig1, self.ax1 = plt.subplots(1,1)
        # cax = plt.gca()
        # cax.axes.xaxis.set_ticklabels([])
        # self.fig2, self.ax2 = plt.subplots(1,1)
        # cax = plt.gca()
        # cax.axes.xaxis.set_ticklabels([])
        # self.fig3, self.ax3 = plt.subplots(1,1)
        # cax = plt.gca()
        # cax.axes.xaxis.set_ticklabels([])
        # plt.subplots_adjust (hspace = 0.35)
        #
        # self.hrs = [0.0]
        # self.y0 = [0.0]
        # self.y1 = [0.0]
        # self.y2 = [0.0]
        # self.y3 = [0.0]
        # self.y0min = 0
        # self.y0max = 2
        # self.y1min = 0.0
        # self.y1max = 2
        # self.y2min = 0.0
        # self.y2max = 0.0
        # self.y3min = 0.0
        # self.y3max = 0.0
        # self.hour_stop = 4.0
        #
        # self.ln0 = Line2D (self.hrs, self.y0, color='red')
        # self.ln1 = Line2D (self.hrs, self.y1, color='red')
        # self.ln2 = Line2D (self.hrs, self.y2, color='blue')
        # self.ln3 = Line2D (self.hrs, self.y3, color='magenta')
        #
        # self.ax0.add_line (self.ln0)
        # self.ax0.set_ylabel('[pu]')
        # self.ax0.set_title ('vpos8', fontsize=10)
        #
        # self.ax1.add_line (self.ln1)
        # self.ax1.set_ylabel('[pu]')
        # self.ax1.set_title ('vpos1', fontsize=10)
        #
        # self.ax2.add_line (self.ln2)
        # self.ax2.set_ylabel('[$]')
        # self.ax2.set_title ('LMP8', fontsize=10)
        #
        # self.ax3.add_line (self.ln3)
        # self.ax3.set_ylabel('[kW]')
        # self.ax3.set_title ('distribution_load8', fontsize=10)
        #
        # self.ax3.set_xlabel('Hours')
        #
        # self.canvas0 = FigureCanvasTkAgg(self.fig0, self.frame)
        # self.canvas0.draw()
        # self.canvas0.get_tk_widget().grid(row=0, sticky = tk.W + tk.E + tk.N + tk.S)
        #
        # self.canvas1 = FigureCanvasTkAgg(self.fig1, self.frame)
        # self.canvas1.draw()
        # self.canvas1.get_tk_widget().grid(row=1, sticky = tk.W + tk.E + tk.N + tk.S)
        #
        # self.canvas2 = FigureCanvasTkAgg(self.fig2, self.frame)
        # self.canvas2.draw()
        # self.canvas2.get_tk_widget().grid(row=2, sticky = tk.W + tk.E + tk.N + tk.S)
        #
        # self.canvas3 = FigureCanvasTkAgg(self.fig3, self.frame)
        # self.canvas3.draw()
        # self.canvas3.get_tk_widget().grid(row=3, sticky = tk.W + tk.E + tk.N + tk.S)
        self.bFNCSactive = False
        self.canvas.config(width=self.frameInCanvas.winfo_width(), height=self.frameInCanvas.winfo_height())

        # topics = fncs.get_keyss()
        # for topic in topics:
        #   pass

    def onFrameConfigure(self, event):
        """Reset the scroll region to encompass the inner frame
        """
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def on_closing(self):
        """Verify whether the user wants to stop TESP simulations before exiting the monitor

        This monitor is itself a FNCS federate, so it can not be shut down without shutting
        down all other FNCS federates in the TESP simulation.
        """
        if messagebox.askokcancel('Quit', 'Do you want to close this window? This is likely to stop all simulations.'):
            self.root.quit()
            self.root.destroy()
            if self.bFNCSactive:
                fncs.finalize()
                self.bFNCSactive = False

    def Quit(self):
        """Shuts down this monitor, and also shuts down FNCS if active
        """
        self.root.quit()
        self.root.destroy()
        if self.bFNCSactive:
            fncs.finalize()
            self.bFNCSactive = False

    def OpenConfig(self):
        """Read the JSON configuration file for this monitor, and initialize the plot axes
        """ 
        fname = filedialog.askopenfilename(initialdir='.',
                                           initialfile='tesp_monitor_ercot.json',
                                           title='Open JSON Monitor Configuration',
                                           filetypes=(('JSON files', '*.json'), ('all files', '*.*')),
                                           defaultextension='json')
        lp = open(fname)
        cfg = json.loads(lp.read())
        self.commands = cfg['commands']
        # for row in self.commands:
        #     args = row['args']
        #     if "fncs_broker" in args:
        #         for en in row['env']:
        #             if "FNCS_CONFIG_FILE" in en:
        #                 self.fncsyaml = en[1]
        #                 break
        #         else:
        #             continue
        #         break
        self.fncsyaml = os.environ['FNCS_CONFIG_FILE']
        if os.path.isfile(self.fncsyaml):
            with open(self.fncsyaml, 'r') as stream:
                try:
                    dd = yaml.load(stream)['values'].items()
                    self.topicDict = dict((v['topic'], k) for k, v in dd)
                except yaml.YAMLError as ex:
                    print(ex)
        else:
            print('could not open FNCS_CONFIG_FILE for fncs')
        topics = list(self.topicDict.keys())
        self.plot0.topicDict = self.topicDict
        self.plot0.combobox.config(values=topics)
        self.plot0.listOfTopics = topics
        self.plot0.title.set(topics[0])
        self.plot1.topicDict = self.topicDict
        self.plot1.combobox.config(values=topics)
        self.plot1.listOfTopics = topics
        self.plot1.title.set(topics[0])
        self.plot2.topicDict = self.topicDict
        self.plot2.combobox.config(values=topics)
        self.plot2.listOfTopics = topics
        self.plot2.title.set(topics[0])
        self.plot3.topicDict = self.topicDict
        self.plot3.combobox.config(values=topics)
        self.plot3.listOfTopics = topics
        self.plot3.title.set(topics[0])
        # convert seconds to minutes
        self.time_stop = cfg['time_stop']
        self.yaml_delta = cfg['yaml_delta']
        self.hour_stop = float(self.time_stop / 3600.0)
        dirpath = os.path.dirname(fname)
        os.chdir(dirpath)
        self.labelvar.set(dirpath)

        self.plot0.ax.set_xlim(0.0, self.hour_stop)
        self.plot1.ax.set_xlim(0.0, self.hour_stop)
        self.plot2.ax.set_xlim(0.0, self.hour_stop)
        self.plot3.ax.set_xlim(0.0, self.hour_stop)
        # self.plot0.ax.set_ylim(0.9, 1.1)
        # self.plot1.ax.set_ylim(0.9, 1.1)
        self.plot0.canvas.draw()
        self.plot1.canvas.draw()
        self.plot2.canvas.draw()
        self.plot3.canvas.draw()

    def kill_all(self):
        """Shut down all FNCS federates in TESP, except for this monitor
        """
        for proc in self.pids:
            print('trying to kill', proc.pid)
            proc.terminate()
        self.root.update()
        print('trying to finalize FNCS')
        if self.bFNCSactive:
            fncs.finalize()
            self.bFNCSactive = False
        print('FNCS finalized')

    def update_plots(self, i):
        """This function is called by Matplotlib for each animation frame

        Each time called, collect FNCS messages until the next time to plot
        has been reached. Then update the plot quantities and return the
        Line2D objects that have been updated for plotting. Check for new
        data outside the plotted vertical range, which triggers a full
        re-draw of the axes. On the last frame, finalize FNCS.

        Args:
          i (int): the animation frame number
        """
        print('.', end='')
        #    print ('frame', i, 'of', self.nsteps)
        bRedraw = False
        while self.time_granted < self.time_stop:  # time in minutes
            self.time_granted = fncs.time_request(self.time_stop)
            events = fncs.get_events()
            self.root.update()
            idx = int(self.time_granted / self.yaml_delta)
            if idx <= self.idxlast:
                continue
            self.idxlast = idx

            v0 = 0.0
            v1 = 0.0
            v2 = 0.0
            v3 = 0.0

            # for topic in events:
            #   value = fncs.get_value(topic)
            #   # if topic == 'vpos8':
            #   #   v1 = float (value.strip('+ degFkW')) / 199185.0
            #   if topic == 'vpos1':
            #     v1 = float(value.strip('+ degFkW')) / 199185.0
            #   elif topic == 'distribution_load8':
            #     v3 = helpers.parse_kw (value)
            #   elif topic == 'vpos8':
            #     v0 = float (value.strip('+ degFkW')) / 199185.0
            #   # elif topic == 'clear_price':
            #   #   v2 = float (value.strip('+ degFkW'))
            #   elif topic == 'LMP8':
            #     v2 = float (value.strip('+ degFkW'))
            #   # elif topic == 'SUBSTATION7':
            #   #   v1 = float (value.strip('+ degFkW')) # already in kW

            for topic in events:
                value = fncs.get_value(topic)
                if topic == self.topicDict[self.plot0.title.get()]:
                    v0 = self.CalculateValue(topic, value, self.plot0)
                elif topic == self.topicDict[self.plot1.title.get()]:
                    v1 = self.CalculateValue(topic, value, self.plot1)
                elif topic == self.topicDict[self.plot2.title.get()]:
                    v2 = self.CalculateValue(topic, value, self.plot2)
                elif topic == self.topicDict[self.plot3.title.get()]:
                    v3 = self.CalculateValue(topic, value, self.plot3)

            retval = [self.plot0.ln, self.plot1.ln, self.plot2.ln, self.plot3.ln]

            h = float(self.time_granted / 3600.0)
            self.plot0.hrs.append(h)
            self.plot1.hrs.append(h)
            self.plot2.hrs.append(h)
            self.plot3.hrs.append(h)
            self.plot0.y.append(v0)  # Vpu
            self.plot1.y.append(v1)  # school kW
            self.plot2.y.append(v2)  # price
            self.plot3.y.append(v3)  # feeder load
            self.plot0.ln.set_data(self.plot0.hrs, self.plot0.y)
            self.plot1.ln.set_data(self.plot1.hrs, self.plot1.y)
            self.plot2.ln.set_data(self.plot2.hrs, self.plot2.y)
            self.plot3.ln.set_data(self.plot3.hrs, self.plot3.y)
            if len(self.plot0.y) == 2:
                diff = abs(self.plot0.y[0] - self.plot0.y[1])
                if diff == 0:
                    self.plot0.ymin = self.plot0.y[0] - 0.00001
                    self.plot0.ymax = self.plot0.y[0] + 0.00001
                else:
                    self.plot0.ymin = min(self.plot0.y) - 0.00001 * diff
                    self.plot0.ymax = max(self.plot0.y) + 0.00001 * diff
                if self.plot0.ymin < 0:
                    self.plot0.ymin = 0
                self.plot0.ax.set_ylim(self.plot0.ymin, self.plot0.ymax)
                self.plot0.fig.canvas.draw()
                diff = abs(self.plot1.y[0] - self.plot1.y[1])
                if diff == 0:
                    self.plot1.ymin = self.plot1.y[0] - 0.00001
                    self.plot1.ymax = self.plot1.y[0] + 0.00001
                else:
                    self.plot1.ymin = min(self.plot1.y) - 0.00001 * diff
                    self.plot1.ymax = max(self.plot1.y) + 0.00001 * diff
                if self.plot1.ymin < 0:
                    self.plot1.ymin = 0
                self.plot1.ax.set_ylim(self.plot1.ymin, self.plot1.ymax)
                self.plot1.fig.canvas.draw()
                diff = abs(self.plot2.y[0] - self.plot2.y[1])
                if diff == 0:
                    self.plot2.ymin = self.plot2.y[0] - 0.00001
                    self.plot2.ymax = self.plot2.y[0] + 0.00001
                else:
                    self.plot2.ymin = min(self.plot2.y) - 0.00001 * diff
                    self.plot2.ymax = max(self.plot2.y) + 0.00001 * diff
                if self.plot2.ymin < 0:
                    self.plot2.ymin = 0
                self.plot2.ax.set_ylim(self.plot2.ymin, self.plot2.ymax)
                self.plot2.fig.canvas.draw()
                diff = abs(self.plot3.y[0] - self.plot3.y[1])
                if diff == 0:
                    self.plot3.ymin = self.plot3.y[0] - 0.00001
                    self.plot3.ymax = self.plot3.y[0] + 0.00001
                else:
                    self.plot3.ymin = min(self.plot3.y) - 0.00001 * diff
                    self.plot3.ymax = max(self.plot3.y) + 0.00001 * diff
                if self.plot3.ymin < 0:
                    self.plot3.ymin = 0
                self.plot3.ax.set_ylim(self.plot3.ymin, self.plot3.ymax)
                self.plot3.fig.canvas.draw()
            else:
                if v0 < self.plot0.ymin or v0 > self.plot0.ymax:
                    if v0 < self.plot0.ymin:
                        self.plot0.ymin = v0
                    if v0 > self.plot0.ymax:
                        self.plot0.ymax = v0
                    self.plot0.ax.set_ylim(self.plot0.ymin, self.plot0.ymax)
                    self.plot0.fig.canvas.draw()
                if v1 < self.plot1.ymin or v1 > self.plot1.ymax:
                    if v1 < self.plot1.ymin:
                        self.plot1.ymin = v1
                    if v1 > self.plot1.ymax:
                        self.plot1.ymax = v1
                    self.plot1.ax.set_ylim(self.plot1.ymin, self.plot1.ymax)
                    self.plot1.fig.canvas.draw()
                if v2 < self.plot2.ymin or v2 > self.plot2.ymax:
                    if v2 < self.plot2.ymin:
                        self.plot2.ymin = v2
                    if v2 > self.plot2.ymax:
                        self.plot2.ymax = v2
                    self.plot2.ax.set_ylim(self.plot2.ymin, self.plot2.ymax)
                    self.plot2.fig.canvas.draw()
                if v3 < self.plot3.ymin or v3 > self.plot3.ymax:
                    if v3 < self.plot3.ymin:
                        self.plot3.ymin = v3
                    if v3 > self.plot3.ymax:
                        self.plot3.ymax = v3
                    self.plot3.ax.set_ylim(self.plot3.ymin, self.plot3.ymax)
                    self.plot3.fig.canvas.draw()

            # if bRedraw:
            #   self.plot0.fig.canvas.draw()
            #   self.plot1.fig.canvas.draw()
            #   self.plot2.fig.canvas.draw()
            #   self.plot3.fig.canvas.draw()
            if i >= (self.nsteps - 1):
                fncs.finalize()
                self.bFNCSactive = False
            return retval

    def CalculateValue(self, topic, value, plot):
        """Parses a value from FNCS to plot

        Args:
            topic (str): the FNCS topic
            value (str): the FNCS value
            plot (ChoosablePlot): the plot that will be updated; contains the voltage base if needed

        Returns:
            float: the parsed value
        """
        if 'vpos' in topic:
            return float(value.strip('+ degFkW')) / float(plot.voltageBase.get())
        elif 'distribution_load' in topic:
            return helpers.parse_kw(value)
        elif 'airtemp' in topic:
            return float(value.strip('+ degFkW'))
        elif 'LMP' in topic:
            return float(value.strip('+ degFkW'))
        else:
            return float(value.strip('+ degFkW'))

    def launch_all(self):
        """Launches the simulators, initializes FNCS and starts the animated plots
        """
        self.root.update()

        print('launching all simulators')
        self.pids = []
        for row in self.commands:
            procargs = row['args']
            procenv = os.environ.copy()
            for var in row['env']:
                procenv[var[0]] = var[1]
            logfd = None
            if 'log' in row:
                logfd = open(row['log'], 'w')
            proc = subprocess.Popen(procargs, env=procenv, stdout=logfd)
            self.pids.append(proc)

        print('launched', len(self.pids), 'simulators')
        self.root.update()

        fncs.initialize()
        self.bFNCSactive = True
        print('FNCS initialized')
        self.nsteps = int(self.time_stop / self.yaml_delta)
        self.idxlast = -1
        self.time_granted = 0

        ani = animation.FuncAnimation(self.plot0.fig, self.update_plots, frames=self.nsteps,
                                      blit=True, repeat=False, interval=0)
        ani = animation.FuncAnimation(self.plot1.fig, self.update_plots, frames=self.nsteps,
                                      blit=True, repeat=False, interval=0)
        ani = animation.FuncAnimation(self.plot2.fig, self.update_plots, frames=self.nsteps,
                                      blit=True, repeat=False, interval=0)
        ani = animation.FuncAnimation(self.plot3.fig, self.update_plots, frames=self.nsteps,
                                      blit=True, repeat=False, interval=0)
        self.plot0.fig.canvas.draw()
        self.plot1.fig.canvas.draw()
        self.plot2.fig.canvas.draw()
        self.plot3.fig.canvas.draw()


#    fncs.finalize()

def show_tesp_monitor():
    """Creates and displays the monitor GUI
    """
    root = tk.Tk()
    root.title('Transactive Energy Simulation Platform: Solution Monitor')
    my_gui = TespMonitorGUI(root)
    root.update()
    while True:
        try:
            root.mainloop()
            break
        except UnicodeDecodeError:
            pass


class ChoosablePlot(tk.Frame):
    """Hosts one Matplotlib animation with a selected variable to plot

    Args:
        master:
        color (str): Matplotlib color of the line to plot
        xLabel (str): label for the x axis
        topicDict (dict): dictionary of FNCS topic choices to plot
        kwargs: arbitrary keyword arguments
    Attributes:
        topicDict:
        listOfTopics:
        root (Tk): the TCL Tk toolkit instance
        fig (Figure): animated Matplotlib figure hosted on the GUI
        ax (Axes): set of 4 xy axes to plot on
        xLabel (str): horizontal axis label
        yLabel (str): vertical axis label
        hrs:
        y:
        ymin (float):
        ymax (float):
        color:
        title (str):
        ln (Line2D):
        topicSelectionRow:
        topicLabel:
        combobox:
        voltageLabel:
        voltageBase:
        voltageBaseTextbox:
    """
    def __init__(self, master, color='red', xLabel='', topicDict={}, **kwargs):
        super().__init__(master, background='#ffffff', **kwargs)
        self.root = master
        self.fig, self.ax = plt.subplots(1, 1, figsize=(9, 1.8))
        self.topicDict = topicDict
        self.listOfTopics = list(topicDict.keys())
        self.xLabel = xLabel
        self.yLabel = ''

        self.hrs = []
        self.y = []
        self.ymin = 0
        self.ymax = 0
        self.color = color
        self.title = tk.StringVar(self)
        if self.listOfTopics:
            self.title.set(self.listOfTopics[0])

        self.ln = Line2D(self.hrs, self.y, color=color)
        self.ax.add_line(self.ln)
        self.ax.set_ylabel('[LMP]')
        self.ax.set_title(self.title.get(), fontsize=10)

        if self.xLabel == '':
            cax = plt.gca()
            cax.axes.xaxis.set_ticklabels([])
        else:
            self.ax.set_xlabel(self.xLabel)
            plt.gcf().set_size_inches(9, 2.5)
            plt.gcf().subplots_adjust(bottom=0.35)
            plt.gca().set_autoscale_on(False)
        # self.rowconfigure (0, weight=0)
        # self.columnconfigure (0, weight=1)
        # self.columnconfigure (1, weight=10)
        # self.columnconfigure (2, weight=2)
        # self.columnconfigure (3, weight=10)
        self.topicSelectionRow = tk.Frame(self, background='#ffffff')
        self.topicSelectionRow.grid(row=0, sticky='w', padx=(75, 0))

        self.topicLabel = tk.Label(self.topicSelectionRow, text="Topic: ", background='#ffffff')
        self.topicLabel.pack(side='left')

        self.combobox = ttk.Combobox(self.topicSelectionRow, values=self.listOfTopics, textvariable=self.title,
                                     background="#ffffff", width=50, state="readonly")
        self.combobox.pack(side='left')
        self.combobox.bind("<<ComboboxSelected>>", self.onTopicSelected)

        self.voltageLabel = tk.Label(self.topicSelectionRow, text="Voltage Base: ", background='#ffffff')
        self.voltageLabel.pack(side='left', padx=(30, 0))

        self.voltageBase = tk.StringVar(self)
        self.voltageBase.set('199185')
        self.voltageBaseTextbox = tk.Entry(self.topicSelectionRow, textvariable=self.voltageBase, relief=tk.RIDGE,
                                           background="#ffffff", width=15, highlightthickness=0,
                                           highlightbackground='White')
        self.voltageBaseTextbox.pack(side='left')

        self.voltageLabel.pack_forget()
        self.voltageBaseTextbox.pack_forget()

        self.canvas = FigureCanvasTkAgg(self.fig, self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1)

    def onTopicSelected(self, event):
        """ Change the GUI labels to match selected plot variable

        Args:
            event:
        """
        selectedTopic = self.title.get()
        self.ax.set_title(selectedTopic, fontsize=10)
        if 'vpos' in self.topicDict[selectedTopic]:
            self.voltageLabel.pack(side='left')
            self.voltageBaseTextbox.pack(side='left')
            self.ax.set_ylabel('[pu]')
        else:
            self.voltageLabel.pack_forget()
            self.voltageBaseTextbox.pack_forget()
            if 'distribution_load' in self.topicDict[selectedTopic]:
                self.ax.set_ylabel('[kW]')
            elif 'airtemp' in self.topicDict[selectedTopic]:
                self.ax.set_ylabel('[degF]')
            elif 'LMP' in self.topicDict[selectedTopic]:
                self.ax.set_ylabel('[$]')
            else:
                self.ax.set_ylabel('[$]')
        self.canvas.draw()



# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: tesp_case.py
"""Creates and fills a subdirectory with files to run a TESP simulation

Use *tesp_config* to graphically edit the case configuration

Public Functions:
    :make_tesp_case: sets up for a single-shot TESP case
    :make_monte_carlo_cases: sets up for a Monte Carlo TESP case of up to 20 shots
    :first_tesp_feeder: customization of make_tesp_case that will accept more feeders
    :add_tesp_feeder: add another feeder to the case directory created by first_tesp_feeder
"""
import sys
import json
import subprocess
import os
import stat
import shutil
from datetime import datetime
import tesp_support.helpers as helpers
import tesp_support.make_ems as idf

if sys.platform == 'win32':
    pycall = 'python'
else:
    pycall = 'python3'

def write_tesp_case (config, cfgfile, freshdir = True):
    """Writes the TESP case from data structure to JSON file

    This function assumes one GridLAB-D, one EnergyPlus, one PYPOWER
    and one substation_loop federate will participate in the TESP simulation.
    See the DSO+T study functions, which are customized to ERCOT 8-bus and 
    200-bus models, for examples of other configurations.

    The TESP support directories, working directory and case name are all specified 
    in *config*. This function will create one directory as follows:

        * workdir = config['SimulationConfig']['WorkingDirectory']
        * casename = config['SimulationConfig']['CaseName']
        * new directory created will be *casedir* = workdir/casename

    This function will read or copy several files that are specified in the *config*.
    They should all exist. These include taxonomy feeders, GridLAB-D schedules, 
    weather files, a base EnergyPlus model, a base PYPOWER model, and supporting scripts
    for the end user to invoke from the *casedir*.  The user could add more base model files
    weather files or schedule files under the TESP support directory, where this *tesp_case*
    module will be able to find and use them.

    This function will launch and wait for 6 subprocesses to assist in the 
    case configuration. All must execute successfully:

        * TMY3toTMY2_ansi, which converts the user-selected TMY3 file to TMY2
        * tesp.convert_tmy2_to_epw, which converts the TMY2 file to EPW for EnergyPlus
        * tesp.TMY3toCSV, which converts the TMY3 file to CSV for the weather agent
        * tesp.populate_feeder, which populates the user-selected taxonomy feeder with houses and DER
        * tesp.glm_dict, which creates metadata for the populated feeder
        * tesp.prep_substation, which creates metadata and FNCS configurations for the substation agents

    As the configuration process finishes, several files are written to *casedir*:

        * Casename.glm: the GridLAB-D model, copied and modified from the TESP support directory
        * Casename_FNCS_Config.txt: FNCS subscriptions and publications, included by Casename.glm
        * Casename_agent_dict.json: metadata for the simple_auction and hvac agents
        * Casename_glm_dict.json: metadata for Casename.glm
        * Casename_pp.json: the PYPOWER model, copied and modified from the TESP support directory
        * Casename_substation.yaml: FNCS subscriptions and time step for the substation, which manages the simple_auction and hvac controllers
        * NonGLDLoad.txt: non-responsive load data for the PYPOWER model buses, currently hard-wired for the 9-bus model. See the ERCOT case files for examples of expanded options.
        * SchoolDualController.idf: the EnergyPlus model, copied and modified from the TESP support directory 
        * WA-Yakima_Air_Terminal.epw: the selected weather file for EnergyPlus, others can be selected
        * WA-Yakima_Air_Terminal.tmy3: the selected weather file for GridLAB-D, others can be selected
        * appliance_schedules.glm: time schedules for GridLAB-D
        * clean.sh: Linux/Mac OS X helper to clean up simulation outputs
        * commercial_schedules.glm: non-responsive non-responsive time schedules for GridLAB-D, invariant
        * eplus.yaml: FNCS subscriptions and time step for EnergyPlus
        * eplus_agent.yaml: FNCS subscriptions and time step for the EnergyPlus agent
        * kill5570.sh: Linux/Mac OS X helper to kill all federates listening on port 5570
        * launch_auction.py: helper script for the GUI solution monitor to launch the substation federate
        * launch_pp.py: helper script for the GUI solution monitor to launch the PYPOWER federate
        * monitor.py: helper to launch the GUI solution monitor (FNCS_CONFIG_FILE envar must be set for this process, see gui.sh under examples/te30)
        * plots.py: helper script that will plot a selection of case outputs
        * pypower.yaml: FNCS subscriptions and time step for PYPOWER
        * run.sh: Linux/Mac OS X helper to launch the TESP simulation
        * tesp_monitor.json: shell commands and other configuration data for the solution monitor GUI
        * tesp_monitor.yaml: FNCS subscriptions and time step for the solution monitor GUI
        * water_and_setpoint_schedule_v5.glm: non-responsive time schedules for GridLAB-D, invariant
        * weather.dat: CSV file of temperature, pressure, humidity, solar direct, solar diffuse and wind speed

    Args:
        config (dict): the complete case data structure
        cfgfile (str): the name of the JSON file that was read
        freshdir (boolean): flag to create the directory and base files anew

    Todo:
        * Write gui.sh, per the te30 examples
    """
    tespdir = os.path.expandvars (os.path.expanduser (config['SimulationConfig']['SourceDirectory']))
    feederdir = tespdir + '/feeders/'
    scheduledir = tespdir + '/schedules/'
    weatherdir = tespdir + '/weather/'
    eplusdir = tespdir + '/energyplus/'
    ppdir = tespdir + '/pypower/'
    miscdir = tespdir + '/misc/'
    print ('feeder backbone files from', feederdir)
    print ('schedule files from', scheduledir)
    print ('weather files from', weatherdir)
    print ('E+ files from', eplusdir)
    print ('pypower backbone files from', ppdir)

    casename = config['SimulationConfig']['CaseName']
    workdir = config['SimulationConfig']['WorkingDirectory']
    if len(workdir) > 2:
        casedir = workdir
    else:
        casedir = workdir + casename
    glmroot = config['BackboneFiles']['TaxonomyChoice']
    print ('case files written to', casedir)

    if freshdir == True:
        if os.path.exists(casedir):
            shutil.rmtree(casedir)
        os.makedirs(casedir)

    StartTime = config['SimulationConfig']['StartTime']
    EndTime = config['SimulationConfig']['EndTime']
    time_fmt = '%Y-%m-%d %H:%M:%S'
    dt1 = datetime.strptime (StartTime, time_fmt)
    dt2 = datetime.strptime (EndTime, time_fmt)
    seconds = int ((dt2 - dt1).total_seconds())
    days = seconds / 86400
    WeatherYear = dt1.year
    print ('run', days, 'days or', seconds, 'seconds in weather year', WeatherYear)

    (rootweather, weatherext) = os.path.splitext(config['WeatherPrep']['DataSource'])
    EpRef = config['EplusConfiguration']['ReferencePrice']
    EpRamp = config['EplusConfiguration']['Slope']
    EpLimHi = config['EplusConfiguration']['OffsetLimitHi']
    EpLimLo = config['EplusConfiguration']['OffsetLimitLo']
    EpWeather = rootweather + '.epw' # config['EplusConfiguration']['EnergyPlusWeather']
    EpStepsPerHour = int (config['EplusConfiguration']['StepsPerHour'])
    EpBuilding = config['EplusConfiguration']['BuildingChoice']
    EpEMS = config['EplusConfiguration']['EMSFile']
    EpXfmrKva = config['EplusConfiguration']['EnergyPlusXfmrKva']
    EpVolts = config['EplusConfiguration']['EnergyPlusServiceV']
    EpBus = config['EplusConfiguration']['EnergyPlusBus']
    EpMetricsKey = EpBuilding # os.path.splitext (EpFile)[0]
    EpAgentStop = str (seconds) + 's'
    EpStep = int (60 / EpStepsPerHour) # minutes
    EpAgentStep = str(int (60 / EpStepsPerHour)) + 'm'
    EpMetricsFile = 'eplus_' + casename + '_metrics.json'
    GldFile = casename + '.glm'
    GldMetricsFile = casename + '_metrics.json'
    AgentDictFile = casename + '_agent_dict.json'
    PPJsonFile = casename + '_pp.json'
    SubstationYamlFile = casename + '_substation.yaml'
    WeatherConfigFile = casename + '_FNCS_Weather_Config.json'

    weatherfile = weatherdir + rootweather + '.tmy3'
    eplusfile = eplusdir + EpBuilding + '.idf'
    emsfile = eplusdir + EpEMS + '.idf'
    eplusout = casedir + '/Merged.idf'
    ppfile = ppdir + config['BackboneFiles']['PYPOWERFile']
    ppcsv = ppdir + config['PYPOWERConfiguration']['CSVLoadFile']

    # copy some boilerplate files
    if freshdir == True:
#        shutil.copy (miscdir + 'clean.bat', casedir)
#        shutil.copy (miscdir + 'kill5570.bat', casedir)
#        shutil.copy (miscdir + 'killold.bat', casedir)
#        shutil.copy (miscdir + 'list5570.bat', casedir)
        shutil.copy (miscdir + 'clean.sh', casedir)
        shutil.copy (miscdir + 'kill5570.sh', casedir)
        shutil.copy (miscdir + 'kill23404.sh', casedir)
        shutil.copy (miscdir + 'killboth.sh', casedir)
        shutil.copy (miscdir + 'monitor.py', casedir)
        shutil.copy (miscdir + 'plots.py', casedir)
        shutil.copy (scheduledir + 'appliance_schedules.glm', casedir)
        shutil.copy (scheduledir + 'commercial_schedules.glm', casedir)
        shutil.copy (scheduledir + 'water_and_setpoint_schedule_v5.glm', casedir)
    #    shutil.copy (weatherfile, casedir)
        # process TMY3 ==> weather.dat
        cmdline = pycall + """ -c "import tesp_support.api as tesp;tesp.weathercsv('""" + weatherfile + """','""" + casedir + '/weather.dat' + """','""" + StartTime + """','""" + EndTime + """',""" + str(WeatherYear) + """)" """
        print (cmdline)
    #    quit()
        pw0 = subprocess.Popen (cmdline, shell=True)
        pw0.wait()

    #########################################
    # set up EnergyPlus, if the user wants it
    bUseEplus = False
    if len(EpBus) > 0:
        bUseEplus = True
        idf.merge_idf (eplusfile, emsfile, StartTime, EndTime, eplusout, EpStepsPerHour)

        # process TMY3 ==> TMY2 ==> EPW
        cmdline = 'TMY3toTMY2_ansi ' + weatherfile + ' > ' + casedir + '/' + rootweather + '.tmy2'
        print (cmdline)
        pw1 = subprocess.Popen (cmdline, shell=True)
        pw1.wait()
        cmdline = pycall + """ -c "import tesp_support.api as tesp;tesp.convert_tmy2_to_epw('""" + casedir + '/' + rootweather + """')" """
        print (cmdline)
        pw2 = subprocess.Popen (cmdline, shell=True)
        pw2.wait()
        os.remove (casedir + '/' + rootweather + '.tmy2')

        # write the EnergyPlus YAML files 
        op = open (casedir + '/eplus.yaml', 'w')
        print ('name: eplus', file=op)
        print ('time_delta:', str (EpStep) + 'm', file=op)
        print ('broker: tcp://localhost:5570', file=op)
        print ('values:', file=op)
        print ('    COOL_SETP_DELTA:', file=op)
        print ('        topic: eplus_agent/cooling_setpoint_delta', file=op)
        print ('        default: 0', file=op)
        print ('    HEAT_SETP_DELTA:', file=op)
        print ('        topic: eplus_agent/heating_setpoint_delta', file=op)
        print ('        default: 0', file=op)
        op.close()

        epjyamlstr = """name: eplus_agent
time_delta: """ + str(EpAgentStep) + """
broker: tcp://localhost:5570
values:
    kwhr_price:
        topic: sub1/clear_price
        default: 0.10
    indoor_air:
        topic: eplus/EMS INDOOR AIR TEMPERATURE
        default: 0
    outdoor_air:
        topic: eplus/ENVIRONMENT SITE OUTDOOR AIR DRYBULB TEMPERATURE
        default: 0
    cooling_volume:
        topic: eplus/EMS COOLING VOLUME
        default: 0
    heating_volume:
        topic: eplus/EMS HEATING VOLUME
        default: 0
    cooling_controlled_load:
        topic: eplus/EMS COOLING CONTROLLED LOAD
        default: 0
    cooling_schedule_temperature:
        topic: eplus/EMS COOLING SCHEDULE TEMPERATURE
        default: 0
    cooling_setpoint_temperature:
        topic: eplus/EMS COOLING SETPOINT TEMPERATURE
        default: 0
    cooling_current_temperature:
        topic: eplus/EMS COOLING CURRENT TEMPERATURE
        default: 0
    cooling_power_state:
        topic: eplus/EMS COOLING POWER STATE
        default: 0
    heating_controlled_load:
        topic: eplus/EMS HEATING CONTROLLED LOAD
        default: 0
    heating_schedule_temperature:
        topic: eplus/EMS HEATING SCHEDULE TEMPERATURE
        default: 0
    heating_setpoint_temperature:
        topic: eplus/EMS HEATING SETPOINT TEMPERATURE
        default: 0
    heating_current_temperature:
        topic: eplus/EMS HEATING CURRENT TEMPERATURE
        default: 0
    heating_power_state:
        topic: eplus/EMS HEATING POWER STATE
        default: 0
    electric_demand_power:
        topic: eplus/WHOLE BUILDING FACILITY TOTAL ELECTRIC DEMAND POWER
        default: 0
    hvac_demand_power:
        topic: eplus/WHOLE BUILDING FACILITY TOTAL HVAC ELECTRIC DEMAND POWER
        default: 0
    ashrae_uncomfortable_hours:
        topic: eplus/FACILITY FACILITY THERMAL COMFORT ASHRAE 55 SIMPLE MODEL SUMMER OR WINTER CLOTHES NOT COMFORTABLE TIME
        default: 0
    occupants_total:
        topic: eplus/EMS OCCUPANT COUNT
        default: 0
"""
        op = open (casedir + '/eplus_agent.yaml', 'w')
        print (epjyamlstr, file=op)
        op.close()

        epaSubs = []
        epaSubs.append ({"name": "kwhr_price","key": "sub1/clear_price","type": "double"})
        epaPubs = []
        epaPubs.append ({"global":False, "key":"power_A", "type":"double", "unit":"W"})
        epaPubs.append ({"global":False, "key":"power_B", "type":"double", "unit":"W"})
        epaPubs.append ({"global":False, "key":"power_C", "type":"double", "unit":"W"})
        epaPubs.append ({"global":False, "key":"bill_mode", "type":"string"})
        epaPubs.append ({"global":False, "key":"price", "type":"double", "unit":"$/kwh"})
        epaPubs.append ({"global":False, "key":"monthly_fee", "type":"double", "unit":"$"})
        epaConfig = {}
        epaConfig["name"] = "eplus_agent"
        epaConfig["log_level"] = 4
        epaConfig["period"] = 60 * EpStep
        epaConfig["subscriptions"] = epaSubs
        epaConfig["publications"] = epaPubs
        op = open (casedir + '/helics_eplus_agent.json', 'w', encoding='utf-8')
        json.dump (epaConfig, op, ensure_ascii=False, indent=2)
        op.close()

    ###################################
    # dynamically import the base PYPOWER case
    import importlib.util
    spec = importlib.util.spec_from_file_location('ppbasecase', ppfile)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    ppcase = mod.ppcasefile()
    #print (ppcase)

    # make ppcase JSON serializable
    ppcase['bus'] = ppcase['bus'].tolist()
    ppcase['gen'] = ppcase['gen'].tolist()
    ppcase['branch'] = ppcase['branch'].tolist()
    ppcase['areas'] = ppcase['areas'].tolist()
    ppcase['gencost'] = ppcase['gencost'].tolist()
    ppcase['DSO'] = ppcase['DSO'].tolist()
    ppcase['UnitsOut'] = ppcase['UnitsOut'].tolist()
    ppcase['BranchesOut'] = ppcase['BranchesOut'].tolist()

    # update the case from config JSON
    ppcase['StartTime'] = config['SimulationConfig']['StartTime']
    ppcase['Tmax'] = int(seconds)
    ppcase['Period'] = config['AgentPrep']['MarketClearingPeriod']
    ppcase['dt'] = config['PYPOWERConfiguration']['PFStep']
    ppcase['CSVFile'] = config['PYPOWERConfiguration']['CSVLoadFile']
    if config['PYPOWERConfiguration']['ACOPF'] == 'AC':
        ppcase['opf_dc'] = 0
    else:
        ppcase['opf_dc'] = 1
    if config['PYPOWERConfiguration']['ACPF'] == 'AC':
        ppcase['pf_dc'] = 0
    else:
        ppcase['pf_dc'] = 1
    fncsBus = int (config['PYPOWERConfiguration']['GLDBus'])
    fncsScale = float (config['PYPOWERConfiguration']['GLDScale'])
    ppcase['DSO'][0][0] = fncsBus
    ppcase['DSO'][0][2] = fncsScale
    baseKV = float(config['PYPOWERConfiguration']['TransmissionVoltage'])
    for row in ppcase['bus']:
        if row[0] == fncsBus:
            row[9] = baseKV

    if len(config['PYPOWERConfiguration']['UnitOutStart']) > 0 and len(config['PYPOWERConfiguration']['UnitOutEnd']) > 0:
        dt3 = datetime.strptime (config['PYPOWERConfiguration']['UnitOutStart'], time_fmt)
        tout_start = int ((dt3 - dt1).total_seconds())
        dt3 = datetime.strptime (config['PYPOWERConfiguration']['UnitOutEnd'], time_fmt)
        tout_end = int ((dt3 - dt1).total_seconds())
        ppcase['UnitsOut'][0] = [int(config['PYPOWERConfiguration']['UnitOut']), tout_start, tout_end]
    else:
        ppcase['UnitsOut'] = []

    if len(config['PYPOWERConfiguration']['BranchOutStart']) > 0 and len(config['PYPOWERConfiguration']['BranchOutEnd']) > 0:
        dt3 = datetime.strptime (config['PYPOWERConfiguration']['BranchOutStart'], time_fmt)
        tout_start = int ((dt3 - dt1).total_seconds())
        dt3 = datetime.strptime (config['PYPOWERConfiguration']['BranchOutEnd'], time_fmt)
        tout_end = int ((dt3 - dt1).total_seconds())
        ppcase['BranchesOut'][0] = [int(config['PYPOWERConfiguration']['BranchOut']), tout_start, tout_end]
    else:
        ppcase['BranchesOut'] = []

    fp = open (casedir + '/' + casename + '_pp.json', 'w')
    json.dump (ppcase, fp, indent=2)
    fp.close ()

    ppyamlstr = """name: pypower
time_delta: """ + str(config['PYPOWERConfiguration']['PFStep']) + """s
broker: tcp://localhost:5570
values:
    SUBSTATION7:
        topic: gld1/distribution_load
        default: 0
    UNRESPONSIVE_MW:
        topic: sub1/unresponsive_mw
        default: 0
    RESPONSIVE_MAX_MW:
        topic: sub1/responsive_max_mw
        default: 0
    RESPONSIVE_C2:
        topic: sub1/responsive_c2
        default: 0
    RESPONSIVE_C1:
        topic: sub1/responsive_c1
        default: 0
    RESPONSIVE_DEG:
        topic: sub1/responsive_deg
        default: 0
"""
    if freshdir == True:
        shutil.copy (ppcsv, casedir)
        op = open (casedir + '/pypower.yaml', 'w')
        print (ppyamlstr, file=op)
        op.close()
        ppSubs = []
        ppSubs.append ({"name": "SUBSTATION7","key": "gld1/distribution_load","type": "complex"})
        ppSubs.append ({"name": "UNRESPONSIVE_MW","key": "sub1/unresponsive_mw","type": "double"})
        ppSubs.append ({"name": "RESPONSIVE_MAX_MW","key": "sub1/responsive_max_mw","type": "double"})
        ppSubs.append ({"name": "RESPONSIVE_C1","key": "sub1/responsive_c1","type": "double"})
        ppSubs.append ({"name": "RESPONSIVE_C2","key": "sub1/responsive_c2","type": "double"})
        ppSubs.append ({"name": "RESPONSIVE_DEG","key": "sub1/responsive_deg","type": "integer"})
        ppPubs = []
        ppPubs.append ({"global":False, "key":"three_phase_voltage_B7", "type":"double"})
        ppPubs.append ({"global":False, "key":"LMP_B7", "type":"double"})
        ppConfig = {}
        ppConfig["name"] = "pypower"
        ppConfig["log_level"] = 4
        ppConfig["period"] = int (config['PYPOWERConfiguration']['PFStep'])
        ppConfig["subscriptions"] = ppSubs
        ppConfig["publications"] = ppPubs
        op = open (casedir + '/pypowerConfig.json', 'w', encoding='utf-8')
        json.dump (ppConfig, op, ensure_ascii=False, indent=2)
        op.close()

    # write a YAML for the solution monitor
    tespyamlstr = """name = tesp_monitor
time_delta = """ + str(int(config['AgentPrep']['MarketClearingPeriod'])) + """s
broker: tcp://localhost:5570
aggregate_sub: true
values:
  vpos7:
    topic: pypower/three_phase_voltage_B7
    default: 0
    type: double
    list: false
  LMP7:
    topic: pypower/LMP_B7
    default: 0
    type: double
    list: false
  clear_price:
    topic: sub1/clear_price
    default: 0
    type: double
    list: false
  distribution_load:
    topic: gld1/distribution_load
    default: 0
    type: complex
    list: false
  power_A:
    topic: eplus_agent/power_A
    default: 0
    type: double
    list: false
  electric_demand_power:
    topic: eplus/WHOLE BUILDING FACILITY TOTAL ELECTRIC DEMAND POWER
    default: 0
    type: double
    list: false
"""
    if freshdir == True:
        op = open (casedir + '/tesp_monitor.yaml', 'w')
        print (tespyamlstr, file=op)
        op.close()

    cmdline = pycall + """ -c "import tesp_support.api as tesp;tesp.populate_feeder('""" + cfgfile + """')" """
    print (cmdline)
    p1 = subprocess.Popen (cmdline, shell=True)
    p1.wait()
    glmfile = casedir + '/' + casename

    cmdline = pycall + """ -c "import tesp_support.api as tesp;tesp.glm_dict('""" + glmfile + """')" """
    print (cmdline)
    p2 = subprocess.Popen (cmdline, shell=True)
    p2.wait()

    cmdline = pycall + """ -c "import tesp_support.api as tesp;tesp.prep_substation('""" + glmfile + """','""" + cfgfile + """')" """
    print (cmdline)
    p3 = subprocess.Popen (cmdline, shell=True)
    p3.wait()

    if freshdir == False:
        return

    # FNCS shell scripts and chmod for Mac/Linux - need to specify python3
    aucline = """python3 -c "import tesp_support.api as tesp;tesp.substation_loop('""" + AgentDictFile + """','""" + casename + """')" """
    ppline = """python3 -c "import tesp_support.api as tesp;tesp.pypower_loop('""" + PPJsonFile + """','""" + casename + """')" """
    weatherline = """python3 -c "import tesp_support.api as tesp;tesp.startWeatherAgent('weather.dat')" """

    shfile = casedir + '/run.sh'
    op = open (shfile, 'w')
    if bUseEplus:
        print ('(export FNCS_BROKER="tcp://*:5570" && export FNCS_FATAL=YES && exec fncs_broker 6 &> broker.log &)', file=op)
        print ('(export FNCS_CONFIG_FILE=eplus.yaml && export FNCS_FATAL=YES && exec energyplus -w ' 
               + EpWeather + ' -d output -r Merged.idf &> eplus.log &)', file=op)
        print ('(export FNCS_CONFIG_FILE=eplus_agent.yaml && export FNCS_FATAL=YES && exec eplus_agent', EpAgentStop, EpAgentStep, 
               EpMetricsKey, EpMetricsFile, EpRef, EpRamp, EpLimHi, EpLimLo, '&> eplus_agent.log &)', file=op)
    else:
        print ('(export FNCS_BROKER="tcp://*:5570" && export FNCS_FATAL=YES && exec fncs_broker 4 &> broker.log &)', file=op)
    print ('(export FNCS_FATAL=YES && exec gridlabd -D USE_FNCS -D METRICS_FILE='
           + GldMetricsFile + ' ' + GldFile + ' &> gld1.log &)', file=op)
    print ('(export FNCS_CONFIG_FILE=' + SubstationYamlFile + ' && export FNCS_FATAL=YES && exec ' + aucline + ' &> sub1.log &)', file=op)
    print ('(export FNCS_CONFIG_FILE=pypower.yaml && export FNCS_FATAL=YES && export FNCS_LOG_STDOUT=yes && exec ' + ppline + ' &> pypower.log &)', file=op)
    print ('(export WEATHER_CONFIG=' + WeatherConfigFile + ' && export FNCS_FATAL=YES && export FNCS_LOG_STDOUT=yes && exec ' + weatherline + ' &> weather.log &)', file=op)
    op.close()
    st = os.stat (shfile)
    os.chmod (shfile, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    shfile = casedir + '/kill5570.sh'
    os.chmod (shfile, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    shfile = casedir + '/clean.sh'
    os.chmod (shfile, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    # HELICS shell scripts and chmod for Mac/Linux - need to specify python3
    PypowerConfigFile = 'pypowerConfig.json'
    SubstationConfigFile = casename + '_HELICS_substation.json'
    WeatherConfigFile = casename + '_HELICS_Weather_Config.json'
    aucline = """python3 -c "import tesp_support.api as tesp;tesp.substation_loop('""" + AgentDictFile + """','""" + casename + """',helicsConfig='""" + SubstationConfigFile + """')" """
    ppline = """python3 -c "import tesp_support.api as tesp;tesp.pypower_loop('""" + PPJsonFile + """','""" + casename + """',helicsConfig='""" + PypowerConfigFile + """')" """

    shfile = casedir + '/runh.sh'
    op = open (shfile, 'w')
    if bUseEplus:
      print ('# FNCS federation is energyplus with agent', file=op)
      print ('(export FNCS_BROKER="tcp://*:5570" && export FNCS_FATAL=YES && exec fncs_broker 2 &> fncs_broker.log &)', file=op)
      print ('(export FNCS_CONFIG_FILE=eplus.yaml && export FNCS_FATAL=YES && exec energyplus -w ' 
             + EpWeather + ' -d output -r Merged.idf &> fncs_eplus.log &)', file=op)
      print ('(export FNCS_CONFIG_FILE=eplus_agent.yaml && export FNCS_FATAL=YES && exec eplus_agent', EpAgentStop, EpAgentStep, 
             EpMetricsKey, EpMetricsFile, EpRef, EpRamp, EpLimHi, EpLimLo, 'helics_eplus_agent.json &> dual_eplus_agent.log &)', file=op)
      print ('# HELICS federation is GridLAB-D, PYPOWER, substation and weather; the E+ agent was already started as part of the FNCS federation', file=op)
      print ('(exec helics_broker -f 5 --loglevel=4 --name=mainbroker &> helics_broker.log &)', file=op)
    else:
      print ('(exec helics_broker -f 4 --loglevel=4 --name=mainbroker &> helics_broker.log &)', file=op)
    print ('(exec gridlabd -D USE_HELICS -D METRICS_FILE='+ GldMetricsFile + ' ' + GldFile + ' &> helics_gld1.log &)', file=op)
    print ('(exec ' + aucline + ' &> helics_sub1.log &)', file=op)
    print ('(exec ' + ppline + ' &> helics_pypower.log &)', file=op)
    print ('(export WEATHER_CONFIG=' + WeatherConfigFile + ' && exec ' + weatherline + ' &> helics_weather.log &)', file=op)
    op.close()
    st = os.stat (shfile)
    os.chmod (shfile, st.st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    # commands for launching Python federates
    op = open (casedir + '/launch_auction.py', 'w')
    print ('import tesp_support.api as tesp', file=op)
    print ('tesp.substation_loop(\'' + AgentDictFile + '\',\'' + casename + '\')', file=op)
    op.close()
    op = open (casedir + '/launch_pp.py', 'w')
    print ('import tesp_support.api as tesp', file=op)
    print ('tesp.pypower_loop(\'' + PPJsonFile + '\',\'' + casename + '\')', file=op)
    op.close()
    op = open (casedir + '/tesp_monitor.json', 'w')
    cmds = {'time_stop':seconds, 
            'yaml_delta':int(config['AgentPrep']['MarketClearingPeriod']), 
            'commands':[]}
    if bUseEplus:
        cmds['commands'].append({'args':['fncs_broker', '6'], 
                           'env':[['FNCS_BROKER', 'tcp://*:5570'],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                           'log':'broker.log'})
        cmds['commands'].append({'args':['EnergyPlus', '-w', EpWeather, '-d', 'output', '-r', 'Merged.idf'], 
                           'env':[['FNCS_CONFIG_FILE', 'eplus.yaml'],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                           'log':'eplus.log'})
        cmds['commands'].append({'args':['eplus_agent', EpAgentStop, EpAgentStep, EpMetricsKey, EpMetricsFile, EpRef, EpRamp, EpLimHi, EpLimLo], 
                           'env':[['FNCS_CONFIG_FILE', 'eplus_agent.yaml'],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                           'log':'eplus_agent.log'})
    else:
        cmds['commands'].append({'args':['fncs_broker', '6'], 
                           'env':[['FNCS_BROKER', 'tcp://*:5570'],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                           'log':'broker.log'})
    cmds['commands'].append({'args':['gridlabd', '-D', 'USE_FNCS', '-D', 'METRICS_FILE=' + GldMetricsFile, GldFile], 
                       'env':[['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                       'log':'gld1.log'})
    cmds['commands'].append({'args':[pycall, 'launch_auction.py'], 
                       'env':[['FNCS_CONFIG_FILE', SubstationYamlFile],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                       'log':'sub1.log'})
    cmds['commands'].append({'args':[pycall, 'launch_pp.py'], 
                       'env':[['FNCS_CONFIG_FILE', 'pypower.yaml'],['FNCS_FATAL', 'YES'],['FNCS_LOG_STDOUT', 'yes']], 
                       'log':'pypower.log'})
    json.dump (cmds, op, indent=2)
    op.close()

def make_tesp_case (cfgfile = 'test.json'):
    """Wrapper function for a single TESP case configuration.

    This function opens the JSON file, and calls *write_tesp_case*

    Args:
        cfgfile (str): JSON file containing the TESP case configuration
    """
    lp = open (cfgfile).read()
    config = json.loads(lp)
    write_tesp_case (config, cfgfile)

def modify_mc_config (config, mcvar, band, sample):
    """Helper function that modifies the Monte Carlo configuration for a specific sample, i.e., shot

    For variables that have a band associated, the agent preparation code will apply
    additional randomization. This applies to thermostat ramps, offset limits, and
    period starting or ending times. For those variables, the Monte Carlo sample
    value is a mean, and the agent preparation code will apply a uniform distribution
    to obtain the actual value for each house.
    """
    if mcvar == 'ElectricCoolingParticipation':
        config['FeederGenerator'][mcvar] = sample
    elif mcvar == 'ThermostatRampMid':
        config['AgentPrep']['ThermostatRampLo'] = sample - 0.5 * band
        config['AgentPrep']['ThermostatRampHi'] = sample + 0.5 * band
    elif mcvar == 'ThermostatOffsetLimit':
        config['AgentPrep']['ThermostatOffsetLimitLo'] = sample - 0.5 * band
        config['AgentPrep']['ThermostatOffsetLimitHi'] = sample + 0.5 * band
    elif mcvar == 'WeekdayEveningStartMid':
        config['ThermostatSchedule']['WeekdayEveningStartLo'] = sample - 0.5 * band
        config['ThermostatSchedule']['WeekdayEveningStartHi'] = sample + 0.5 * band
    elif mcvar == 'WeekdayEveningSetMid':
        config['ThermostatSchedule']['WeekdayEveningSetLo'] = sample - 0.5 * band
        config['ThermostatSchedule']['WeekdayEveningSetHi'] = sample + 0.5 * band

def make_monte_carlo_cases (cfgfile = 'test.json'):
    """Writes up to 20 TESP simulation case setups to a directory for Monte Carlo simulations

    Latin hypercube sampling is recommended; sample values may be specified via *tesp_config*

    Args:
        cfgfile (str): JSON file containing the TESP case configuration
    """
    lp = open (cfgfile).read()
    config = json.loads(lp)
    mc_cfg = 'monte_carlo_sample_' + cfgfile
    basecase = config['SimulationConfig']['CaseName']

    mc = config['MonteCarloCase']
    n = mc['NumCases']
    var1 = mc['Variable1']
    var2 = mc['Variable2']
    var3 = mc['Variable3']
    band1 = mc['Band1']
    band2 = mc['Band2']
    band3 = mc['Band3']
    samples1 = mc['Samples1']
    samples2 = mc['Samples2']
    samples3 = mc['Samples3']
#    print (var1, var2, var3, n)
    for i in range(n):
        mc_case = basecase + '_' + str(i+1)
        config['SimulationConfig']['CaseName'] = mc_case
        modify_mc_config (config, var1, band1, samples1[i])
        modify_mc_config (config, var2, band2, samples2[i])
        modify_mc_config (config, var3, band3, samples3[i])
        op = open (mc_cfg, 'w')
        print (json.dumps(config), file=op)
        op.close()
#        print (mc_case, mc['Samples1'][i], mc['Samples2'][i], mc['Samples3'][i])
        write_tesp_case (config, mc_cfg)

def add_tesp_feeder (cfgfile):
    """Wrapper function to start a single TESP case configuration.

    This function opens the JSON file, and calls *write_tesp_case* for just the
    GridLAB-D files. The subdirectory *targetdir* doesn't have to match the 
    case name in *cfgfile*, and it should be created first with *make_tesp_case*

    Args:
        cfgfile (str): JSON file containing the TESP case configuration
        targetdir (str): directory, based on cwd, to receive the TESP case files
    """
    print ('additional TESP feeder from', cfgfile)
    lp = open (cfgfile).read()
    config = json.loads(lp)
    write_tesp_case (config, cfgfile, freshdir = False)




# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: substation.py
"""Manages the simple_auction and hvac agents for the te30 and sgip1 examples

Public Functions:
    :substation_loop: initializes and runs the agents

Todo:
    * Getting an overflow error when killing process - investigate whether that happens if simulation runs to completion
    * Allow changes in the starting date and time; now it's always midnight on July 1, 2013
    * Allow multiple markets per substation, e.g., 5-minute and day-ahead for the DSO+T study

"""
import sys
try:
  import helics
except:
  pass
try:
  import tesp_support.fncs as fncs
except:
  pass
import tesp_support.simple_auction as auction
import tesp_support.hvac as hvac
import tesp_support.helpers as helpers
import json
import math
from datetime import datetime
from datetime import timedelta
#import gc
#import cProfile
#import pstats
if sys.platform != 'win32':
  import resource

def helics_substation_loop (configfile, metrics_root, hour_stop, flag, helicsConfig):
  print ('starting HELICS substation loop', configfile, metrics_root, hour_stop, flag, flush=True)
  print ('##,tnow,tclear,ClearType,ClearQ,ClearP,BuyCount,BuyUnresp,BuyResp,SellCount,SellUnresp,SellResp,MargQ,MargFrac,LMP,RefLoad,ConSurplus,AveConSurplus,SupplierSurplus,UnrespSupplierSurplus', flush=True)
  bWantMarket = True
  if flag == 'NoMarket':
    bWantMarket = False
    print ('Disabled the market', flush=True)
  time_stop = int (hour_stop * 3600) # simulation time in seconds
  StartTime = '2013-07-01 00:00:00 -0800'
  time_fmt = '%Y-%m-%d %H:%M:%S %z'
  dt_now = datetime.strptime (StartTime, time_fmt)

  # ====== load the JSON dictionary; create the corresponding objects =========

  lp = open (configfile).read()
  dict = json.loads(lp)

  market_key = list(dict['markets'].keys())[0]  # only using the first market
  market_row = dict['markets'][market_key]
  unit = market_row['unit']

  auction_meta = {'clearing_price':{'units':'USD','index':0},'clearing_type':{'units':'[0..5]=[Null,Fail,Price,Exact,Seller,Buyer]','index':1},'consumer_surplus':{'units':'USD','index':2},'average_consumer_surplus':{'units':'USD','index':3},'supplier_surplus':{'units':'USD','index':4}}
  controller_meta = {'bid_price':{'units':'USD','index':0},'bid_quantity':{'units':unit,'index':1}}
  auction_metrics = {'Metadata':auction_meta,'StartTime':StartTime}
  controller_metrics = {'Metadata':controller_meta,'StartTime':StartTime}

  aucObj = auction.simple_auction (market_row, market_key)

  dt = float(dict['dt'])
  period = aucObj.period

  # Initialize controllers, map HELICS values to Python attributes
  subTemp = {}
  subVolt = {}
  subState = {}
  subHVAC = {} 
  pubMtrMode = {}
  pubMtrPrice = {}
  pubMtrMonthly = {}
  pubHeating = {}
  pubCooling = {}
  pubDeadband = {}
  hFed = helics.helicsCreateValueFederateFromConfig(helicsConfig)
  pubCount = helics.helicsFederateGetPublicationCount(hFed)
  subCount = helics.helicsFederateGetInputCount(hFed)
# for i in range(pubCount):
#   pub = helics.helicsFederateGetPublicationByIndex(hFed, i)
#   key = helics.helicsPublicationGetKey (pub)
#   print ('** Available HELICS publication key', i, key)
# for i in range(subCount):
#   sub = helics.helicsFederateGetInputByIndex(hFed, i)
#   key = helics.helicsInputGetKey(sub)
#   target = helics.helicsSubscriptionGetKey(sub)
#   print ('== Available HELICS subscription key', i, key, 'target', target)
  gldName = dict['GridLABD']
  fedName = helics.helicsFederateGetName(hFed)
  bulkName = 'pypower'

  subFeeder = helics.helicsFederateGetSubscription (hFed, gldName + '/distribution_load')
  subLMP = helics.helicsFederateGetSubscription (hFed, bulkName + '/LMP_B7')
  pubC1 = helics.helicsFederateGetPublication (hFed, fedName + '/responsive_c1')
  pubC2 = helics.helicsFederateGetPublication (hFed, fedName + '/responsive_c2')
  pubDeg = helics.helicsFederateGetPublication (hFed, fedName + '/responsive_deg')
  pubMax = helics.helicsFederateGetPublication (hFed, fedName + '/responsive_max_mw')
  pubUnresp = helics.helicsFederateGetPublication (hFed, fedName + '/unresponsive_mw')
  pubAucPrice = helics.helicsFederateGetPublication (hFed, fedName + '/clear_price')

  hvacObjs = {}
  hvac_keys = list(dict['controllers'].keys())
  for key in hvac_keys:
    row = dict['controllers'][key]
    hvacObjs[key] = hvac.hvac (row, key, aucObj)
    ctl = hvacObjs[key]
    hseSubTopic = gldName + '/' + ctl.houseName
    mtrSubTopic = gldName + '/' + ctl.meterName
    mtrPubTopic = fedName + '/' + ctl.meterName
    ctlPubTopic = fedName + '/' + ctl.name
#    print ('{:s} hseSub={:s} mtrSub={:s}  mtrSub={:s}  ctlPub={:s}'.format (key, hseSubTopic, mtrSubTopic, mtrPubTopic, ctlPubTopic))
    subTemp[ctl] = helics.helicsFederateGetSubscription (hFed, hseSubTopic + '#air_temperature')
    subVolt[ctl] = helics.helicsFederateGetSubscription (hFed, mtrSubTopic + '#measured_voltage_1')
    subState[ctl] = helics.helicsFederateGetSubscription (hFed, hseSubTopic + '#power_state')
    subHVAC[ctl] = helics.helicsFederateGetSubscription (hFed, hseSubTopic + '#hvac_load')
    pubMtrMode[ctl] = helics.helicsFederateGetPublication (hFed, mtrPubTopic + '/bill_mode')
    pubMtrPrice[ctl] = helics.helicsFederateGetPublication (hFed, mtrPubTopic + '/price')
    pubMtrMonthly[ctl] = helics.helicsFederateGetPublication (hFed, mtrPubTopic + '/monthly_fee')
    pubHeating[ctl] = helics.helicsFederateGetPublication (hFed, ctlPubTopic + '/heating_setpoint')
    pubCooling[ctl] = helics.helicsFederateGetPublication (hFed, ctlPubTopic + '/cooling_setpoint')
    pubDeadband[ctl] = helics.helicsFederateGetPublication (hFed, ctlPubTopic + '/thermostat_deadband')

  helics.helicsFederateEnterExecutingMode(hFed)

  aucObj.initAuction()
  LMP = aucObj.mean
  refload = 0.0
  bSetDefaults = True

  tnext_bid = period - 2 * dt  #3 * dt  # controllers calculate their final bids
  tnext_agg = period - 2 * dt  # auction calculates and publishes aggregate bid
  tnext_opf = period - 1 * dt  # PYPOWER executes OPF and publishes LMP (no action here)
  tnext_clear = period         # clear the market with LMP
  tnext_adjust = period        # + dt   # controllers adjust setpoints based on their bid and clearing

  time_granted = 0
  time_last = 0
  while (time_granted < time_stop):
    nextHELICSTime = int(min ([tnext_bid, tnext_agg, tnext_clear, tnext_adjust, time_stop]))
#    fncs.update_time_delta (nextFNCSTime-time_granted)
    time_granted = int (helics.helicsFederateRequestTime(hFed, nextHELICSTime))
    time_delta = time_granted - time_last
    time_last = time_granted
    hour_of_day = 24.0 * ((float(time_granted) / 86400.0) % 1.0)
#    print (dt_now, time_delta, timedelta (seconds=time_delta))
    dt_now = dt_now + timedelta (seconds=time_delta)
    day_of_week = dt_now.weekday()
    hour_of_day = dt_now.hour
#    print ('STEP', time_last, time_granted, time_stop, time_delta, hour_of_day, day_of_week, tnext_bid, tnext_agg, tnext_opf, tnext_clear, tnext_adjust, flush=True)
    LMP = helics.helicsInputGetDouble (subLMP)
    aucObj.set_lmp (LMP)
    refload = 0.001 * helics.helicsInputGetDouble (subFeeder)  # supposed to be kW?
    aucObj.set_refload (refload)
#    print ('  REFLOAD', refload, 'LMP', LMP, flush=True)
    for key, obj in hvacObjs.items():
      obj.air_temp = helics.helicsInputGetDouble (subTemp[obj])
      val = helics.helicsInputGetComplex (subVolt[obj])
      obj.mtr_v = math.sqrt (val[0] * val[0] + val[1] * val[1])
      obj.hvac_load = helics.helicsInputGetDouble (subHVAC[obj])
      val = helics.helicsInputGetString (subState[obj])
      if val == 'OFF':
        obj.hvac_on = False
      else:
        obj.hvac_on = True
#    print ('  HVAC INPUTS', flush=True)

    # set the time-of-day schedule
    for key, obj in hvacObjs.items():
      if obj.change_basepoint (hour_of_day, day_of_week):
        helics.helicsPublicationPublishDouble (pubCooling[obj], obj.basepoint)
    if bSetDefaults:
      for key, obj in hvacObjs.items():
        helics.helicsPublicationPublishString (pubMtrMode[obj], 'HOURLY')
        helics.helicsPublicationPublishDouble (pubMtrMonthly[obj], 0.0)
        helics.helicsPublicationPublishDouble (pubDeadband[obj], obj.deadband)
        helics.helicsPublicationPublishDouble (pubHeating[obj], 60.0)
      bSetDefaults = False
#      print ('  SET DEFAULTS', flush=True)

    if time_granted >= tnext_bid:
      aucObj.clear_bids()
      time_key = str (int (tnext_clear))
      controller_metrics [time_key] = {}
      for key, obj in hvacObjs.items():
        bid = obj.formulate_bid () # bid is [price, quantity, on_state]
        if bid is not None:
          if bWantMarket:
            aucObj.collect_bid (bid)
          controller_metrics[time_key][obj.name] = [bid[0], bid[1]]
      tnext_bid += period
#      print ('  COLLECT BIDS', flush=True)

    if time_granted >= tnext_agg:
      aucObj.aggregate_bids()
      helics.helicsPublicationPublishDouble (pubUnresp, aucObj.agg_unresp)
      helics.helicsPublicationPublishDouble (pubMax, aucObj.agg_resp_max)
      helics.helicsPublicationPublishDouble (pubC2, aucObj.agg_c2)
      helics.helicsPublicationPublishDouble (pubC1, aucObj.agg_c1)
      helics.helicsPublicationPublishInteger (pubDeg, aucObj.agg_deg)
      tnext_agg += period
#      print ('  AGGREGATE BIDS', flush=True)

    if time_granted >= tnext_clear:
      if bWantMarket:
        aucObj.clear_market(tnext_clear, time_granted)
        aucObj.surplusCalculation(tnext_clear, time_granted)
        helics.helicsPublicationPublishDouble (pubAucPrice, aucObj.clearing_price)
        for key, obj in hvacObjs.items():
          obj.inform_bid (aucObj.clearing_price)
      time_key = str (int (tnext_clear))
      auction_metrics [time_key] = {aucObj.name:[aucObj.clearing_price, aucObj.clearing_type, aucObj.consumerSurplus, aucObj.averageConsumerSurplus, aucObj.supplierSurplus]}
      tnext_clear += period
#      print ('  CLEARED MARKET', flush=True)

    if time_granted >= tnext_adjust:
      if bWantMarket:
        for key, obj in hvacObjs.items():
          helics.helicsPublicationPublishDouble (pubMtrPrice[obj], aucObj.clearing_price)
          if obj.bid_accepted ():
            helics.helicsPublicationPublishDouble (pubCooling[obj], obj.setpoint)
      tnext_adjust += period
#      print ('  ADJUSTED', flush=True)

  # ==================== Finalize the metrics output ===========================

  print ('writing metrics', flush=True)
  auction_op = open ('auction_' + metrics_root + '_metrics.json', 'w')
  controller_op = open ('controller_' + metrics_root + '_metrics.json', 'w')
  print (json.dumps(auction_metrics), file=auction_op)
  print (json.dumps(controller_metrics), file=controller_op)
  auction_op.close()
  controller_op.close()
  helpers.stop_helics_federate (hFed)

def fncs_substation_loop (configfile, metrics_root, hour_stop=48, flag='WithMarket'):
    """Helper function that initializes and runs the agents

    Reads configfile. Writes *auction_metrics_root_metrics.json* and
    *controller_metrics_root_metrics.json* upon completion.

    Args:
        configfile (str): fully qualified path to the JSON agent configuration file
        metrics_root (str): base name of the case for metrics output
        hour_stop (float): number of hours to simulation
        flag (str): WithMarket or NoMarket to use the simple_auction, or not
    """
    print ('starting FNCS substation loop', configfile, metrics_root, hour_stop, flag, flush=True)
    print ('##,tnow,tclear,ClearType,ClearQ,ClearP,BuyCount,BuyUnresp,BuyResp,SellCount,SellUnresp,SellResp,MargQ,MargFrac,LMP,RefLoad,ConSurplus,AveConSurplus,SupplierSurplus,UnrespSupplierSurplus', flush=True)
    bWantMarket = True
    if flag == 'NoMarket':
        bWantMarket = False
        print ('Disabled the market', flush=True)
    time_stop = int (hour_stop * 3600) # simulation time in seconds
    StartTime = '2013-07-01 00:00:00 -0800'
    time_fmt = '%Y-%m-%d %H:%M:%S %z'
    dt_now = datetime.strptime (StartTime, time_fmt)

    # ====== load the JSON dictionary; create the corresponding objects =========

    lp = open (configfile).read()
    dict = json.loads(lp)

    market_key = list(dict['markets'].keys())[0]  # only using the first market
    market_row = dict['markets'][market_key]
    unit = market_row['unit']

    auction_meta = {'clearing_price':{'units':'USD','index':0},'clearing_type':{'units':'[0..5]=[Null,Fail,Price,Exact,Seller,Buyer]','index':1},'consumer_surplus':{'units':'USD','index':2},'average_consumer_surplus':{'units':'USD','index':3},'supplier_surplus':{'units':'USD','index':4}}
    controller_meta = {'bid_price':{'units':'USD','index':0},'bid_quantity':{'units':unit,'index':1}}
    auction_metrics = {'Metadata':auction_meta,'StartTime':StartTime}
    controller_metrics = {'Metadata':controller_meta,'StartTime':StartTime}

    aucObj = auction.simple_auction (market_row, market_key)

    dt = float(dict['dt'])
    period = aucObj.period

    topicMap = {} # to dispatch incoming FNCS messages; 0..5 for LMP, Feeder load, airtemp, mtr volts, hvac load, hvac state
    topicMap['LMP'] = [aucObj, 0]
    topicMap['refload'] = [aucObj, 1]

    hvacObjs = {}
    hvac_keys = list(dict['controllers'].keys())
    for key in hvac_keys:
        row = dict['controllers'][key]
        hvacObjs[key] = hvac.hvac (row, key, aucObj)
        ctl = hvacObjs[key]
        topicMap[key + '#Tair'] = [ctl, 2]
        topicMap[key + '#V1'] = [ctl, 3]
        topicMap[key + '#Load'] = [ctl, 4]
        topicMap[key + '#On'] = [ctl, 5]

    # ==================== Time step looping under FNCS ===========================

    fncs.initialize()
    aucObj.initAuction()
    LMP = aucObj.mean
    refload = 0.0
    bSetDefaults = True

    tnext_bid = period - 2 * dt  #3 * dt  # controllers calculate their final bids
    tnext_agg = period - 2 * dt  # auction calculates and publishes aggregate bid
    tnext_opf = period - 1 * dt  # PYPOWER executes OPF and publishes LMP (no action here)
    tnext_clear = period         # clear the market with LMP
    tnext_adjust = period        # + dt   # controllers adjust setpoints based on their bid and clearing

    time_granted = 0
    time_last = 0
    while (time_granted < time_stop):
        nextFNCSTime = int(min ([tnext_bid, tnext_agg, tnext_clear, tnext_adjust, time_stop]))
        fncs.update_time_delta (nextFNCSTime-time_granted)
        time_granted = fncs.time_request (nextFNCSTime)
        time_delta = time_granted - time_last
        time_last = time_granted
        hour_of_day = 24.0 * ((float(time_granted) / 86400.0) % 1.0)
#        print (dt_now, time_delta, timedelta (seconds=time_delta))
        dt_now = dt_now + timedelta (seconds=time_delta)
        day_of_week = dt_now.weekday()
        hour_of_day = dt_now.hour
#        print ('  ', time_last, time_granted, time_stop, time_delta, hour_of_day, day_of_week, flush=True)
        # update the data from FNCS messages
        events = fncs.get_events()
        for topic in events:
            value = fncs.get_value(topic)
            row = topicMap[topic]
            if row[1] == 0:
                LMP = helpers.parse_fncs_magnitude (value)
                aucObj.set_lmp (LMP)
            elif row[1] == 1:
                refload = helpers.parse_kw (value)
                aucObj.set_refload (refload)
            elif row[1] == 2:
                row[0].set_air_temp_from_fncs_str (value)
            elif row[1] == 3:
                row[0].set_voltage_from_fncs_str (value)
            elif row[1] == 4:
                row[0].set_hvac_load_from_fncs_str (value)
            elif row[1] == 5:
                row[0].set_hvac_state_from_fncs_str (value)

        # set the time-of-day schedule
        for key, obj in hvacObjs.items():
            if obj.change_basepoint (hour_of_day, day_of_week):
                fncs.publish (obj.name + '/cooling_setpoint', obj.basepoint)
        if bSetDefaults:
            for key, obj in hvacObjs.items():
                fncs.publish (obj.name + '/bill_mode', 'HOURLY')
                fncs.publish (obj.name + '/monthly_fee', 0.0)
                fncs.publish (obj.name + '/thermostat_deadband', obj.deadband)
                fncs.publish (obj.name + '/heating_setpoint', 60.0)
            bSetDefaults = False

        if time_granted >= tnext_bid:
            aucObj.clear_bids()
            time_key = str (int (tnext_clear))
            controller_metrics [time_key] = {}
            for key, obj in hvacObjs.items():
                bid = obj.formulate_bid () # bid is [price, quantity, on_state]
                if bid is not None:
                    if bWantMarket:
                        aucObj.collect_bid (bid)
                    controller_metrics[time_key][obj.name] = [bid[0], bid[1]]
            tnext_bid += period

        if time_granted >= tnext_agg:
            aucObj.aggregate_bids()
            fncs.publish ('unresponsive_mw', aucObj.agg_unresp)
            fncs.publish ('responsive_max_mw', aucObj.agg_resp_max)
            fncs.publish ('responsive_c2', aucObj.agg_c2)
            fncs.publish ('responsive_c1', aucObj.agg_c1)
            fncs.publish ('responsive_deg', aucObj.agg_deg)
            tnext_agg += period

        if time_granted >= tnext_clear:
            if bWantMarket:
                aucObj.clear_market(tnext_clear, time_granted)
                aucObj.surplusCalculation(tnext_clear, time_granted)
                fncs.publish ('clear_price', aucObj.clearing_price)
                for key, obj in hvacObjs.items():
                    obj.inform_bid (aucObj.clearing_price)
            time_key = str (int (tnext_clear))
            auction_metrics [time_key] = {aucObj.name:[aucObj.clearing_price, aucObj.clearing_type, aucObj.consumerSurplus, aucObj.averageConsumerSurplus, aucObj.supplierSurplus]}
            tnext_clear += period
#            print ('garbage collecting at', time_granted, 'finds', gc.collect(), 'unreachable objects', flush=True)

        if time_granted >= tnext_adjust:
            if bWantMarket:
                for key, obj in hvacObjs.items():
                    fncs.publish (obj.name + '/price', aucObj.clearing_price)
                    if obj.bid_accepted ():
                        fncs.publish (obj.name + '/cooling_setpoint', obj.setpoint)
            tnext_adjust += period

    # ==================== Finalize the metrics output ===========================

    print ('writing metrics', flush=True)
    auction_op = open ('auction_' + metrics_root + '_metrics.json', 'w')
    controller_op = open ('controller_' + metrics_root + '_metrics.json', 'w')
    print (json.dumps(auction_metrics), file=auction_op)
    print (json.dumps(controller_metrics), file=controller_op)
    auction_op.close()
    controller_op.close()

    print ('finalizing FNCS', flush=True)
    fncs.finalize()

def substation_loop (configfile, metrics_root, hour_stop=48, flag='WithMarket', helicsConfig=None):
    """Wrapper for *inner_substation_loop*

    When *inner_substation_loop* finishes, timing and memory metrics will be printed
    for non-Windows platforms.
    """
    if helicsConfig is not None:
      helics_substation_loop (configfile, metrics_root, hour_stop, flag, helicsConfig)
    else:
      fncs_substation_loop (configfile, metrics_root, hour_stop, flag)

#    gc.enable() 
#    gc.set_debug(gc.DEBUG_LEAK) 

#    profiler = cProfile.Profile ()
#    args = (configfile, metrics_root, hour_stop, flag)
#    profiler.runcall (inner_substation_loop, *args)
#    stats = pstats.Stats(profiler)
#    stats.strip_dirs()
#    stats.sort_stats('cumulative')
#    stats.print_stats()

#    print (gc.collect (), 'unreachable objects')
#    for x in gc.garbage:
#        s = str(x) 
#        print (type(x), ':', len(s), flush=True)
    if sys.platform != 'win32':
        usage = resource.getrusage(resource.RUSAGE_SELF)
        RESOURCES = [
            ('ru_utime', 'User time'),
            ('ru_stime', 'System time'),
            ('ru_maxrss', 'Max. Resident Set Size'),
            ('ru_ixrss', 'Shared Memory Size'),
            ('ru_idrss', 'Unshared Memory Size'),
            ('ru_isrss', 'Stack Size'),
            ('ru_inblock', 'Block inputs'),
            ('ru_oublock', 'Block outputs')]
        print('Resource usage:')
        for name, desc in RESOURCES:
            print('  {:<25} ({:<10}) = {}'.format(desc, name, getattr(usage, name)))
 

if __name__ == '__main__':
    substation_loop('C:\\Users\\wang690\\Desktop\\projects\\TESP\\tesp_1st\\ercot\\case8\\Bus1_agent_dict.json','Bus1',24)


import csv;
import sys;
import numpy as np;
import matplotlib as mpl;
import matplotlib.pyplot as plt;

casefiles = [['SGIP1a','red'],
						 ['SGIP1b','blue'],
						 ['SGIP1c','green'],
						 ['SGIP1d','magenta'],
						 ['SGIP1e','cyan'],
						 ['SGIP1ex','orange']]

# here are the columns, after ignore 'True/False' with usecols
plotvars = [[0, 't[s]', 'seconds'],
						[1, 'Pload', 'MW'],
						[2, 'P7 (csv)', 'MW'],
						[3, 'GLD Unresp', 'MW'],
						[4, 'P7 (opf)', 'MW'],
						[5, 'Resp (opf)', 'MW'],
						[6, 'GLD Pub', 'MW'],
						[7, 'P7 Min', 'MW'],
						[8, 'V7', 'pu'],
						[9, 'LMP_P7', '$/MWHR'],
						[10, 'LMP_Q7', '$/MWHR'],
						[11, 'Pgen1', 'MW'],
						[12, 'Pgen2', 'MW'],
						[13, 'Pgen3', 'MW'],
						[14, 'Pgen4', 'MW'],
						[15, 'Pdisp', 'MW'],
						[16, 'gencost2', '$'],
						[17, 'gencost1', '$'],
						[18, 'gencost0', '$']]

def MakePlotData(root, idx):
	d1 = np.loadtxt(root + '.csv', skiprows=1, delimiter=',',
									usecols=[0,2,3,4,5,6,7,9,10,11,12,13,14,15,16,17,18,19,20])
	hrs = d1[:,0] / 3600.0

	return hrs, d1[:,idx]

def compare_csv (idx):
	# display a plot - indexed to plotvars
	tmin = 0.0
	tmax = 48.0
	xticks = [0,6,12,18,24,30,36,42,48]

	fig, ax = plt.subplots(1, 1, sharex = 'col')

	for root in casefiles:
		print ('Processing', root[0])
		hrs, vals = MakePlotData(root[0],idx)
		ax.plot(hrs, vals, color=root[1], label=root[0])

	ax.set_title (plotvars[idx][1])
	ax.set_ylabel (plotvars[idx][2])
	ax.set_xlabel ('Hours')
	ax.grid()
	ax.legend()
	ax.set_xlim(tmin,tmax)
	ax.set_xticks(xticks)

	plt.show()





# Copyright (C) 2017-2020 Battelle Memorial Institute
# file: fncsPYPOWER.py
""" PYPOWER solutions under control of FNCS or HELICS for te30 and sgip1 examples

Public Functions:
    :pypower_loop: Initializes and runs the simulation.  
"""
import json
import sys
import warnings
import csv
try:
  import tesp_support.fncs as fncs
except:
  pass
try:
  import helics
except:
  pass
import numpy as np
import pypower.api as pp
from math import sqrt
import math
import re
from copy import deepcopy
import tesp_support.helpers as helpers
#import cProfile
#import pstats
if sys.platform != 'win32':
  import resource

def summarize_opf(res):
  """ Helper function to print optimal power flow solution (debugging)

  Args:
    res (dict): solved PYPOWER case structure
  """
  bus = res['bus']
  gen = res['gen']

  Pload = bus[:,2].sum()
  Pgen = gen[:,1].sum()
  PctLoss = 100.0 * (Pgen - Pload) / Pgen

  print('success =', res['success'], 'in', res['et'], 'seconds')
  print('Total Gen =', Pgen, ' Load =', Pload, ' Loss =', PctLoss, '%')

  print('bus #, Pd, Qd, Vm, LMP_P, LMP_Q, MU_VMAX, MU_VMIN')
  for row in bus:
    print(int(row[0]),row[2],row[3],row[7],row[13],row[14],row[15],row[16])

  print('gen #, bus, Pg, Qg, MU_PMAX, MU_PMIN, MU_PMAX, MU_PMIN')
  idx = 1
  for row in gen:
    print(idx,int(row[0]),row[1],row[2],row[21],row[22],row[23],row[24])
    ++idx

def load_json_case(fname):
  """ Helper function to load PYPOWER case from a JSON file

  Args:
    fname (str): the JSON file to open

  Returns:
    dict: the loaded PYPOWER case structure
  """
  lp = open (fname, encoding='utf-8').read()
  ppc = json.loads(lp)
  ppc['bus'] = np.array (ppc['bus'])
  ppc['gen'] = np.array (ppc['gen'])
  ppc['branch'] = np.array (ppc['branch'])
  ppc['areas'] = np.array (ppc['areas'])
  ppc['gencost'] = np.array (ppc['gencost'])
  ppc['DSO'] = np.array (ppc['DSO'])
  ppc['UnitsOut'] = np.array (ppc['UnitsOut'])
  ppc['BranchesOut'] = np.array (ppc['BranchesOut'])
  return ppc

def make_dictionary(ppc, rootname):
  """ Helper function to write the JSON metafile for post-processing

  Args:
    ppc (dict): PYPOWER case file structure
    rootname (str): to write rootname_m_dict.json
  """
  dsoBuses = {}
  generators = {}
  unitsout = []
  branchesout = []
  bus = ppc['bus']
  gen = ppc['gen']
  cost = ppc['gencost']
  dsoBus = ppc['DSO']
  units = ppc['UnitsOut']
  branches = ppc['BranchesOut']

  for i in range (gen.shape[0]):
    busnum = gen[i,0]
    bustype = bus[busnum-1,1]
    if bustype == 1:
      bustypename = 'pq'
    elif bustype == 2:
      bustypename = 'pv'
    elif bustype == 3:
      bustypename = 'swing'
    else:
      bustypename = 'unknown'
    generators[str(i+1)] = {'bus':int(busnum),'bustype':bustypename,'Pnom':float(gen[i,1]),'Pmax':float(gen[i,8]),'genfuel':'tbd','gentype':'tbd',
      'StartupCost':float(cost[i,1]),'ShutdownCost':float(cost[i,2]), 'c2':float(cost[i,4]), 'c1':float(cost[i,5]), 'c0':float(cost[i,6])}

  for i in range (dsoBus.shape[0]):
    busnum = int(dsoBus[i,0])
    busidx = busnum - 1
    dsoBuses[str(busnum)] = {'Pnom':float(bus[busidx,2]),'Qnom':float(bus[busidx,3]),'area':int(bus[busidx,6]),'zone':int(bus[busidx,10]),
      'ampFactor':float(dsoBus[i,2]),'GLDsubstations':[dsoBus[i,1]]}

  for i in range (units.shape[0]):
    unitsout.append ({'unit':int(units[i,0]),'tout':int(units[i,1]),'tin':int(units[i,2])})

  for i in range (branches.shape[0]):
    branchesout.append ({'branch':int(branches[i,0]),'tout':int(branches[i,1]),'tin':int(branches[i,2])})

  dp = open (rootname + "_m_dict.json", "w")
  ppdict = {'baseMVA':ppc['baseMVA'],'dsoBuses':dsoBuses,'generators':generators,'UnitsOut':unitsout,'BranchesOut':branchesout}
  json.dump (ppdict, dp, ensure_ascii=False, indent=2)
  dp.close()

def parse_mva(arg):
  """ Helper function to parse P+jQ from a FNCS value

  Args:
    arg (str): FNCS value in rectangular format

  Returns:
    float, float: P [MW] and Q [MVAR]
  """
  tok = arg.strip('; MWVAKdrij')
  bLastDigit = False
  bParsed = False
  vals = [0.0,0.0]
  for i in range(len(tok)):
    if tok[i] == '+' or tok[i] == '-':
      if bLastDigit:
        vals[0] = float(tok[:i])
        vals[1] = float(tok[i:])
        bParsed = True
        break
    bLastDigit = tok[i].isdigit()
  if not bParsed:
    vals[0] = float(tok)

  if 'd' in arg:
    vals[1] *= (math.pi / 180.0)
    p = vals[0] * math.cos(vals[1])
    q = vals[0] * math.sin(vals[1])
  elif 'r' in arg:
    p = vals[0] * math.cos(vals[1])
    q = vals[0] * math.sin(vals[1])
  else:
    p = vals[0]
    q = vals[1]

  if 'KVA' in arg:
    p /= 1000.0
    q /= 1000.0
  elif 'MVA' in arg:
    p *= 1.0
    q *= 1.0
  else:  # VA
    p /= 1000000.0
    q /= 1000000.0

  return p, q

def pypower_loop (casefile, rootname, helicsConfig=None):
  """ Public function to start PYPOWER solutions under control of FNCS or HELICS

  The time step, maximum time, and other data must be set up in a JSON file.
  This function will run the case under FNCS, manage the FNCS message traffic,
  and shutdown FNCS upon completion. Five files are written:

  - *rootname.csv*; intermediate solution results during simulation
  - *rootname_m_dict.json*; metadata for post-processing
  - *bus_rootname_metrics.json*; bus metrics for GridLAB-D connections, upon completion
  - *gen_rootname_metrics.json*; bulk system generator metrics, upon completion
  - *sys_rootname_metrics.json*; bulk system-level metrics, upon completion

  Args:
    casefile (str): the configuring JSON file name, without extension
    rootname (str): the root filename for metrics output, without extension
  """

  ppc = load_json_case (casefile)
  StartTime = ppc['StartTime']
  tmax = int(ppc['Tmax'])
  period = int(ppc['Period'])
  dt = int(ppc['dt'])
  make_dictionary (ppc, rootname)

  bus_mp = open ("bus_" + rootname + "_metrics.json", "w")
  gen_mp = open ("gen_" + rootname + "_metrics.json", "w")
  sys_mp = open ("sys_" + rootname + "_metrics.json", "w")
  bus_meta = {'LMP_P':{'units':'USD/kwh','index':0},'LMP_Q':{'units':'USD/kvarh','index':1},
    'PD':{'units':'MW','index':2},'QD':{'units':'MVAR','index':3},'Vang':{'units':'deg','index':4},
    'Vmag':{'units':'pu','index':5},'Vmax':{'units':'pu','index':6},'Vmin':{'units':'pu','index':7}}
  gen_meta = {'Pgen':{'units':'MW','index':0},'Qgen':{'units':'MVAR','index':1},'LMP_P':{'units':'USD/kwh','index':2}}
  sys_meta = {'Ploss':{'units':'MW','index':0},'Converged':{'units':'true/false','index':1}}
  bus_metrics = {'Metadata':bus_meta,'StartTime':StartTime}
  gen_metrics = {'Metadata':gen_meta,'StartTime':StartTime}
  sys_metrics = {'Metadata':sys_meta,'StartTime':StartTime}

  gencost = ppc['gencost']
  dsoBus = ppc['DSO']
  gen = ppc['gen']
  pf_dc = ppc['pf_dc']
  PswingSwitch = 180.0 # if Gen 2 gets close to the max, change swing bus
  if pf_dc > 0:
    PswingSwitch = 191.0
  ppopt_market = pp.ppoption(VERBOSE=0, OUT_ALL=0, PF_DC=ppc['opf_dc'])
  ppopt_regular = pp.ppoption(VERBOSE=0, OUT_ALL=0, PF_DC=ppc['pf_dc'])
  loads = np.loadtxt(ppc['CSVFile'], delimiter=',')

  for row in ppc['UnitsOut']:
    print ('unit  ', row[0], 'off from', row[1], 'to', row[2], flush=True)
  for row in ppc['BranchesOut']:
    print ('branch', row[0], 'out from', row[1], 'to', row[2], flush=True)

  nloads = loads.shape[0]
  ts = 0
  tnext_opf = -dt

  # initializing for metrics collection
  tnext_metrics = 0
  loss_accum = 0
  conv_accum = True
  n_accum = 0
  bus_accum = {}
  gen_accum = {}
  for i in range (dsoBus.shape[0]):
    busnum = int(dsoBus[i,0])
    bus_accum[str(busnum)] = [0,0,0,0,0,0,0,99999.0]
  for i in range (gen.shape[0]):
    gen_accum[str(i+1)] = [0,0,0]

  op = open (rootname + '.csv', 'w')
  print ('t[s],Converged,Pload,P7 (csv),Unresp (opf),P7 (rpf),Resp (opf),GLD Pub,BID?,P7 Min,V7,LMP_P7,LMP_Q7,Pgen1,Pgen2,Pgen3,Pgen4,Pdisp,Deg,c2,c1', file=op, flush=True)

  hFed = None
  pub_lmp = None
  pub_volts = None
  sub_load = None
  sub_unresp = None
  sub_max = None
  sub_c2 = None
  sub_c1 = None
  sub_deg = None
  if helicsConfig is not None:
    hFed = helics.helicsCreateValueFederateFromConfig(helicsConfig)
    fedName = helics.helicsFederateGetName(hFed)
    pubCount = helics.helicsFederateGetPublicationCount(hFed)
    subCount = helics.helicsFederateGetInputCount(hFed)
    for i in range(pubCount):
      pub = helics.helicsFederateGetPublicationByIndex(hFed, i)
      key = helics.helicsPublicationGetKey (pub)
      print ('HELICS publication key', i, key)
      if 'LMP_' in key:
        pub_lmp = pub
      elif 'three_phase_voltage_' in key:
        pub_volts = pub
    for i in range(subCount):
      sub = helics.helicsFederateGetInputByIndex(hFed, i)
      key = helics.helicsInputGetKey(sub)
      target = helics.helicsSubscriptionGetKey(sub)
      print ('HELICS subscription key', i, key, 'target', target)
      upper_target = target.upper() # FNCS-compatible matching
      if 'RESPONSIVE_C2' in upper_target:
        sub_c2 = sub
      if 'RESPONSIVE_C1' in upper_target:
        sub_c1 = sub
      if 'RESPONSIVE_DEG' in upper_target:
        sub_deg = sub
      if 'RESPONSIVE_MAX_MW' in upper_target:
        sub_max = sub
      if 'UNRESPONSIVE_MW' in upper_target:
        sub_unresp = sub
      if 'distribution_load' in target:
        sub_load = sub
    helics.helicsFederateEnterExecutingMode(hFed)
  else:
    fncs.initialize()

  # transactive load components
  csv_load = 0     # from the file
  unresp = 0       # unresponsive load estimate from the auction agent
  resp = 0         # will be the responsive load as dispatched by OPF
  resp_deg = 0     # RESPONSIVE_DEG from DSO
  resp_c1 = 0      # RESPONSIVE_C1 from DSO
  resp_c2 = 0      # RESPONSIVE_C2 from DSO
  resp_max = 0     # RESPONSIVE_MAX_MW from DSO
  feeder_load = 0  # amplified feeder MW

  while ts <= tmax:
    # start by getting the latest inputs from GridLAB-D and the auction
    new_bid = False
    load_scale = float (dsoBus[0][2])
    if hFed is not None: # HELICS inputs, need to be sticky because they were zeroed out just above
      if sub_unresp is not None:
        unresp = helics.helicsInputGetDouble(sub_unresp) * load_scale
        dsoBus[0][3] = unresp # to poke unresponsive estimate into the bus load slot
      if sub_c2 is not None:
        resp_c2 = helics.helicsInputGetDouble(sub_c2) / load_scale
      if sub_c1 is not None:
        resp_c1 = helics.helicsInputGetDouble(sub_c1)
      if sub_deg is not None:
        resp_deg = helics.helicsInputGetInteger(sub_deg)
      if sub_max is not None:
        resp_max = helics.helicsInputGetComplex(sub_max)[0] * load_scale
        if helics.helicsInputIsUpdated(sub_max): #TODO: verify this fires; resp_max updates but new_bid always False to CSV
          new_bid = True
      if sub_load is not None:
        gld_load = helics.helicsInputGetComplex(sub_load)
        feeder_load = gld_load[0] * load_scale / 1.0e6
#      print ('HELICS inputs at', ts, feeder_load, load_scale, unresp, resp_max, resp_c2, resp_c1, resp_deg, new_bid)
    else:  # inputs coming from FNCS
      events = fncs.get_events()
      for topic in events:
        value = fncs.get_value(topic)
        if topic == 'UNRESPONSIVE_MW':
          unresp = load_scale * float(value)
          dsoBus[0][3] = unresp # to poke unresponsive estimate into the bus load slot
          new_bid = True
        elif topic == 'RESPONSIVE_MAX_MW':
          resp_max = load_scale * float(value)
          new_bid = True
        elif topic == 'RESPONSIVE_C2':
          resp_c2 = float(value) / load_scale
          new_bid = True
        elif topic == 'RESPONSIVE_C1':
          resp_c1 = float(value)
          new_bid = True
        elif topic == 'RESPONSIVE_DEG':
          resp_deg = int(value)
          new_bid = True
        else:
          gld_load = parse_mva (value) # actual value, may not match unresp + resp load
          feeder_load = float(gld_load[0]) * load_scale
    if new_bid == True:
      dummy = 2
#      print('**Bid', ts, unresp, resp_max, resp_deg, resp_c2, resp_c1)

    # update the case for bids, outages and CSV loads
    idx = int ((ts + dt) / period) % nloads
    bus = ppc['bus']
    gen = ppc['gen']
    branch = ppc['branch']
    gencost = ppc['gencost']
    csv_load = loads[idx,0]
    bus[4,2] = loads[idx,1]
    bus[8,2] = loads[idx,2]
    # process the generator and branch outages
    for row in ppc['UnitsOut']:
      if ts >= row[1] and ts <= row[2]:
        gen[row[0],7] = 0
      else:
        gen[row[0],7] = 1
    for row in ppc['BranchesOut']:
      if ts >= row[1] and ts <= row[2]:
        branch[row[0],10] = 0
      else:
        branch[row[0],10] = 1

    if resp_deg == 2:
      gencost[4][3] = 3
      gencost[4][4] = -resp_c2
      gencost[4][5] = resp_c1
    elif resp_deg == 1:
      gencost[4][3] = 2
      gencost[4][4] = resp_c1
      gencost[4][5] = 0.0
    else:
      gencost[4][3] = 1
      gencost[4][4] = 999.0
      gencost[4][5] = 0.0
    gencost[4][6] = 0.0

    if ts >= tnext_opf:  # expecting to solve opf one dt before the market clearing period ends, so GridLAB-D has time to use it
      # for OPF, the DSO bus load is CSV + Unresponsive estimate, with Responsive separately dispatchable
      bus = ppc['bus']
      gen = ppc['gen']
      bus[6,2] = csv_load
      for row in ppc['DSO']:
        unresp = float(row[3])
        newidx = int(row[0]) - 1
        if unresp >= feeder_load:
          bus[newidx,2] += unresp
        else:
          bus[newidx,2] += feeder_load
      gen[4][9] = -resp_max
      res = pp.runopf(ppc, ppopt_market)
      if res['success'] == False:
        conv_accum = False
      opf_bus = deepcopy (res['bus'])
      opf_gen = deepcopy (res['gen'])
      lmp = opf_bus[6,13]
      resp = -1.0 * opf_gen[4,1]
      if pub_lmp is not None:
        helics.helicsPublicationPublishDouble(pub_lmp, 0.001 * lmp)
      else:
        fncs.publish('LMP_B7', 0.001 * lmp) # publishing $/kwh
#     print ('  OPF', ts, csv_load, '{:.3f}'.format(unresp), '{:.3f}'.format(resp),
#            '{:.3f}'.format(feeder_load), '{:.3f}'.format(opf_bus[6,2]),
#            '{:.3f}'.format(opf_gen[0,1]), '{:.3f}'.format(opf_gen[1,1]), '{:.3f}'.format(opf_gen[2,1]),
#            '{:.3f}'.format(opf_gen[3,1]), '{:.3f}'.format(opf_gen[4,1]), '{:.3f}'.format(lmp))
      # if unit 2 (the normal swing bus) is dispatched at max, change the swing bus to 9
      if opf_gen[1,1] >= PswingSwitch:
        ppc['bus'][1,1] = 2
        ppc['bus'][8,1] = 3
        print ('  Switching to SWING Bus 9 (Gen 4) at {:d} and {:.2f} MW'.format (ts, opf_gen[1,1]))
      else:
        ppc['bus'][1,1] = 3
        ppc['bus'][8,1] = 1
        print ('  Keeping SWING Bus 2 (Gen 2) at {:d} and {:.2f} MW'.format (ts, opf_gen[1,1]))
      tnext_opf += period
    
    # always update the electrical quantities with a regular power flow
    bus = ppc['bus']
    gen = ppc['gen']
    bus[6,13] = lmp
    gen[0,1] = opf_gen[0, 1]
    gen[1,1] = opf_gen[1, 1]
    gen[2,1] = opf_gen[2, 1]
    gen[3,1] = opf_gen[3, 1]
    # during regular power flow, we use the actual CSV + feeder load, ignore dispatchable load and use actual
    bus[6,2] = csv_load + feeder_load
    gen[4,1] = 0 # opf_gen[4, 1]
    gen[4,9] = 0
    rpf = pp.runpf(ppc, ppopt_regular)
    if rpf[0]['success'] == False:
      conv_accum = False
    bus = rpf[0]['bus']
    gen = rpf[0]['gen']
    
    Pload = bus[:,2].sum()
    Pgen = gen[:,1].sum()
    Ploss = Pgen - Pload

    # update the metrics
    n_accum += 1
    loss_accum += Ploss
    for i in range (dsoBus.shape[0]):
      busnum = int(dsoBus[i,0])
      busidx = busnum - 1
      row = bus[busidx].tolist()
      # LMP_P, LMP_Q, PD, QD, Vang, Vmag, Vmax, Vmin: row[11] and row[12] are Vmax and Vmin constraints
      PD = row[2] + resp # the ERCOT version shows how to track scaled_resp separately for each DSO bus
      Vpu = row[7]
      bus_accum[str(busnum)][0] += row[13]*0.001
      bus_accum[str(busnum)][1] += row[14]*0.001
      bus_accum[str(busnum)][2] += PD
      bus_accum[str(busnum)][3] += row[3]
      bus_accum[str(busnum)][4] += row[8]
      bus_accum[str(busnum)][5] += Vpu
      if Vpu > bus_accum[str(busnum)][6]:
        bus_accum[str(busnum)][6] = Vpu
      if Vpu < bus_accum[str(busnum)][7]:
        bus_accum[str(busnum)][7] = Vpu
    for i in range (gen.shape[0]):
      row = gen[i].tolist()
      busidx = int(row[0] - 1)
      # Pgen, Qgen, LMP_P  (includes the responsive load as dispatched by OPF)
      gen_accum[str(i+1)][0] += row[1]
      gen_accum[str(i+1)][1] += row[2]
      gen_accum[str(i+1)][2] += float(opf_bus[busidx,13])*0.001

    # write the metrics
    if ts >= tnext_metrics:
      sys_metrics[str(ts)] = {rootname:[loss_accum / n_accum,conv_accum]}

      bus_metrics[str(ts)] = {}
      for i in range (dsoBus.shape[0]):
        busnum = int(dsoBus[i,0])
        busidx = busnum - 1
        row = bus[busidx].tolist()
        met = bus_accum[str(busnum)]
        bus_metrics[str(ts)][str(busnum)] = [met[0]/n_accum, met[1]/n_accum, met[2]/n_accum, met[3]/n_accum,
                                             met[4]/n_accum, met[5]/n_accum, met[6], met[7]]
        bus_accum[str(busnum)] = [0,0,0,0,0,0,0,99999.0]

      gen_metrics[str(ts)] = {}
      for i in range (gen.shape[0]):
        met = gen_accum[str(i+1)]
        gen_metrics[str(ts)][str(i+1)] = [met[0]/n_accum, met[1]/n_accum, met[2]/n_accum]
        gen_accum[str(i+1)] = [0,0,0]

      tnext_metrics += period
      n_accum = 0
      loss_accum = 0
      conv_accum = True

    volts = 1000.0 * bus[6,7] * bus[6,9] / sqrt(3.0)  # VLN for GridLAB-D
    if pub_volts is not None:
      helics.helicsPublicationPublishDouble(pub_volts, volts)
    else:
      fncs.publish('three_phase_voltage_B7', volts)

    # CSV file output
    print (ts, res['success'], 
           '{:.3f}'.format(Pload),          # Pload
           '{:.3f}'.format(csv_load),       # P7 (csv)
           '{:.3f}'.format(unresp),         # GLD Unresp
           '{:.3f}'.format(bus[6,2]),       # P7 (rpf)
           '{:.3f}'.format(resp),           # Resp (opf)
           '{:.3f}'.format(feeder_load),    # GLD Pub
           new_bid, 
           '{:.3f}'.format(gen[4,9]),       # P7 Min
           '{:.3f}'.format(bus[6,7]),       # V7
           '{:.3f}'.format(bus[6,13]),      # LMP_P7
           '{:.3f}'.format(bus[6,14]),      # LMP_Q7
           '{:.2f}'.format(gen[0,1]),       # Pgen1
           '{:.2f}'.format(gen[1,1]),       # Pgen2 
           '{:.2f}'.format(gen[2,1]),       # Pgen3
           '{:.2f}'.format(gen[3,1]),       # Pgen4
           '{:.2f}'.format(res['gen'][4, 1]), # Pdisp
           '{:.4f}'.format(resp_deg),       # degree
           '{:.8f}'.format(ppc['gencost'][4, 4]),  # c2
           '{:.8f}'.format(ppc['gencost'][4, 5]),  # c1 
           sep=',', file=op, flush=True)

    # request the next time step, if necessary
    if ts >= tmax:
      print ('breaking out at',ts,flush=True)
      break
    tRequest = min(ts + dt, tmax)
    if hFed is not None:
      ts = int (helics.helicsFederateRequestTime(hFed, tRequest))
    else:
      ts = fncs.time_request(tRequest)

  # ===================================
  print ('writing metrics', flush=True)
  print (json.dumps(sys_metrics), file=sys_mp, flush=True)
  print (json.dumps(bus_metrics), file=bus_mp, flush=True)
  print (json.dumps(gen_metrics), file=gen_mp, flush=True)
  print ('closing files', flush=True)
  bus_mp.close()
  gen_mp.close()
  sys_mp.close()
  op.close()
  if hFed is not None:
    helpers.stop_helics_federate (hFed)
  else:
    print ('finalizing DSO', flush=True)
    fncs.finalize()

  if sys.platform != 'win32':
    usage = resource.getrusage(resource.RUSAGE_SELF)
    RESOURCES = [
      ('ru_utime', 'User time'),
      ('ru_stime', 'System time'),
      ('ru_maxrss', 'Max. Resident Set Size'),
      ('ru_ixrss', 'Shared Memory Size'),
      ('ru_idrss', 'Unshared Memory Size'),
      ('ru_isrss', 'Stack Size'),
      ('ru_inblock', 'Block inputs'),
      ('ru_oublock', 'Block outputs')]
    print('Resource usage:')
    for name, desc in RESOURCES:
      print('  {:<25} ({:<10}) = {}'.format(desc, name, getattr(usage, name)))

# main_loop()
#  profiler = cProfile.Profile ()
#  profiler.runcall (main_loop)
#  stats = pstats.Stats(profiler)
#  stats.strip_dirs()
#  stats.sort_stats('cumulative')
#  stats.print_stats()



# Copyright (C) 2017-2019 Battelle Memorial Institute
# file: tesp_monitor.py
"""Presents a GUI to launch a TESP simulation and monitor its progress

Public Functions:
  :show_tesp_monitor: Initializes and runs the monitor GUI

References:
  `Graphical User Interfaces with Tk <https://docs.python.org/3/library/tk.html>`_

  `Matplotlib Animation <https://matplotlib.org/api/animation_api.html>`_
"""
import sys
import json
import tkinter as tk
import tkinter.ttk as ttk
from tkinter import filedialog
from tkinter import messagebox
import subprocess
import os
try:
  import tesp_support.fncs as fncs
except:
  pass
import tesp_support.helpers as helpers
import time

import numpy as np;
import matplotlib;
matplotlib.use('TkAgg');
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk;
from matplotlib.figure import Figure;
from matplotlib.lines import Line2D
import matplotlib.animation as animation
import matplotlib.pyplot as plt;

class TespMonitorGUI:
  """Manages a GUI with 4 plotted variables, and buttons to stop TESP

  The GUI reads a JSON file with scripted shell commands to launch
  other FNCS federates, and a YAML file with FNCS subscriptions to update
  the solution status. Both JSON and YAML files are written by *tesp.tesp_config*
  The plotted variables provide a sign-of-life and sign-of-stability indication
  for each of the major federates in the te30 or sgip1 examples, namely
  GridLAB-D, PYPOWER, EnergPlus, and the substation_loop that manages a simple_auction
  with multiple hvac agents. If a solution appears to be unstable or must be
  stopped for any other reason, exiting the solution monitor will do so.

  The plots are created and updated with animated and bit-blitted Matplotlib
  graphs hosted on a TkInter GUI. When the JSON and YAML files are loaded,
  the x axis is laid out to match the total TESP simulation time range.

  Args:
    root (Tk): the TCL Tk toolkit instance
    top (Window): the top-level TCL Tk Window
    labelvar (StringVar): used to display the monitor JSON configuration file path
    hrs ([float]): x-axis data array for time in hours, shared by all plots
    y0 ([float]): y-axis data array for PYPOWER bus voltage
    y1 ([float]): y-axis data array for EnergyPlus load
    y2lmp ([float]): y-axis data array for PYPOWER LMP
    y2auc ([float]): y-axis data array for simple_auction cleared_price
    y3fncs ([float]): y-axis data array for GridLAB-D load via FNCS
    y3gld ([float]): y-axis data array for sample-and-hold GridLAB-D load
    gld_load (float): the most recent load published by GridLAB-D; due to the deadband, this value isn't necessary published at every FNCS time step
    y0min (float): the first y axis minimum value
    y0max (float): the first y axis maximum value
    y1min (float): the second y axis minimum value
    y1max (float): the second y axis maximum value
    y2min (float): the third y axis minimum value
    y2max (float): the third y axis maximum value
    y3min (float): the fourth y axis minimum value
    y3max (float): the fourth y axis maximum value
    hour_stop (float): the maximum x axis time value to plot
    ln0 (Line2D): the plotted PYPOWER bus voltage, color GREEN
    ln1 (Line2D): the plotted EnergyPlus load, color RED
    ln2lmp (Line2D): the plotted PYPOWER locational marginal price (LMP), color BLUE
    ln2auc (Line2D): the plotted simple_auction cleared_price, color BLACK
    ln3gld (Line2D): the plotted sample-and-hold GridLAB-D substation load, color MAGENTA
    ln3fncs (Line2D): the plotted GridLAB-D substation load published via FNCS; may be zero if not published for the current animation frame, color CYAN
    fig (Figure): animated Matplotlib figure hosted on the GUI
    ax (Axes): set of 4 xy axes to plot on
    canvas (FigureCanvasTkAgg): a TCL Tk canvas that can host Matplotlib
    bFNCSactive (Boolean): True if a TESP simulation is running with other FNCS federates, False if not
  """
  def __init__ (self, master):
    self.root = master
    self.root.protocol('WM_DELETE_WINDOW', self.on_closing)
    self.top = self.root.winfo_toplevel()
    self.top.rowconfigure (0, weight=1)
    self.top.columnconfigure (0, weight=1)
    self.fig, self.ax = plt.subplots(4,1, sharex = 'col')
    plt.subplots_adjust (hspace = 0.35)

    ttk.Style().configure('TButton', foreground='blue')

    btn = ttk.Button(self.root, text='Open...', command=self.OpenConfig)
    btn.grid(row=0, column=0, sticky=tk.NSEW)
    btn = ttk.Button(self.root, text='Start All', command=self.launch_all)
    btn.grid(row=0, column=1, sticky=tk.NSEW)
    btn = ttk.Button(self.root, text='Kill All', command=self.kill_all)
    btn.grid(row=0, column=2, sticky=tk.NSEW)
    btn = ttk.Button(self.root, text='Quit', command=self.Quit)
    btn.grid(row=0, column=3, sticky=tk.NSEW)
    self.labelvar = tk.StringVar()
    self.labelvar.set('Case')
    lab = ttk.Label(self.root, textvariable=self.labelvar, relief=tk.RIDGE)
    lab.grid(row=0, column=4, sticky=tk.NSEW)

    self.root.rowconfigure (0, weight=0)
    self.root.rowconfigure (1, weight=1)
    self.root.columnconfigure (0, weight=0)
    self.root.columnconfigure (1, weight=0)
    self.root.columnconfigure (2, weight=0)
    self.root.columnconfigure (3, weight=0)
    self.root.columnconfigure (4, weight=1)

    self.hrs = [0.0]
    self.y0 = [1.0]
    self.y1 = [0.0]
    self.y2lmp = [0.0]
    self.y2auc = [0.0]
    self.y3fncs = [0.0]  # GridLAB-D publishes only when changed
    self.gld_load = 0.0
    self.y3gld = [0.0]
    self.y0min = 1.0
    self.y0max = 1.0
    self.y1min = 0.0
    self.y1max = 0.0
    self.y2min = 0.0
    self.y2max = 0.0
    self.y3min = 0.0
    self.y3max = 0.0
    self.hour_stop = 4.0

    self.ln0 = Line2D (self.hrs, self.y0, color='green')
    self.ln1 = Line2D (self.hrs, self.y1, color='red')
    self.ln2auc = Line2D (self.hrs, self.y2auc, color='black')
    self.ln2lmp = Line2D (self.hrs, self.y2lmp, color='blue')
    self.ln3fncs = Line2D (self.hrs, self.y3fncs, color='cyan')
    self.ln3gld = Line2D (self.hrs, self.y3gld, color='magenta')

    self.ax[0].add_line (self.ln0)
    self.ax[0].set_ylabel('[pu]')
    self.ax[0].set_title ('PYPOWER Bus Voltage', fontsize=10)

    self.ax[1].add_line (self.ln1)
    self.ax[1].set_ylabel('[kW]')
    self.ax[1].set_title ('Primary School Load', fontsize=10)

    self.ax[2].add_line (self.ln2auc)
    self.ax[2].add_line (self.ln2lmp)
    self.ax[2].set_ylabel('[$]')
    self.ax[2].set_title ('LMP and Clearing Price', fontsize=10)

    self.ax[3].add_line (self.ln3fncs)
    self.ax[3].add_line (self.ln3gld)
    self.ax[3].set_ylabel('[kW]')
    self.ax[3].set_title ('Total Feeder Load', fontsize=10)

    self.ax[3].set_xlabel('Hours')

    self.canvas = FigureCanvasTkAgg(self.fig, self.root)
    self.canvas.draw()
    self.canvas.get_tk_widget().grid(row=1,columnspan=5, sticky = tk.W + tk.E + tk.N + tk.S)

    self.bFNCSactive = False

  def on_closing(self):
    """Verify whether the user wants to stop TESP simulations before exiting the monitor

    This monitor is itself a FNCS federate, so it can not be shut down without shutting
    down all other FNCS federates in the TESP simulation.
    """
    if messagebox.askokcancel('Quit', 'Do you want to close this window? This is likely to stop all simulations.'):
      self.root.quit()
      self.root.destroy()
      if self.bFNCSactive:
        fncs.finalize()
        self.bFNCSactive = False

  def Quit(self):
    """Shuts down this monitor, and also shuts down FNCS if active
    """
    self.root.quit()
    self.root.destroy()
    if self.bFNCSactive:
      fncs.finalize()
      self.bFNCSactive = False

  def OpenConfig(self):
    """Read the JSON configuration file for this monitor, and initialize the plot axes
    """ 
    fname = filedialog.askopenfilename(initialdir = '.',
                                       initialfile = 'tesp_monitor.json',
                                       title = 'Open JSON Monitor Configuration', 
                                       filetypes = (('JSON files','*.json'),('all files','*.*')),
                                       defaultextension = 'json')
    lp = open (fname)
    cfg = json.loads(lp.read())
    self.commands = cfg['commands']
    # no longer converting seconds to minutes
    self.time_stop = int (cfg['time_stop'])
    self.yaml_delta = int (cfg['yaml_delta'])
    self.hour_stop = float (self.time_stop / 3600.0)
    dirpath = os.path.dirname (fname)
    os.chdir (dirpath)
    self.labelvar.set(dirpath)

    self.ax[0].set_xlim(0.0, self.hour_stop)
    self.ax[1].set_xlim(0.0, self.hour_stop)
    self.ax[2].set_xlim(0.0, self.hour_stop)
    self.ax[3].set_xlim(0.0, self.hour_stop)
    self.ax[0].set_ylim(0.9, 1.1)
    self.fig.canvas.draw()

  def kill_all(self):
    """Shut down all FNCS federates in TESP, except for this monitor
    """
    for proc in self.pids:
      print ('trying to kill', proc.pid)
      proc.terminate()
    self.root.update()
    print ('trying to finalize FNCS')
    if self.bFNCSactive:
      fncs.finalize()
      self.bFNCSactive = False
    print ('FNCS finalized')

  def expand_limits(self, v, vmin, vmax):
    """Whenever a variable meets a vertical axis limit, expand the limits with 10% padding above and below

    Args:
      v (float): the out of range value
      vmin (float): the current minimum vertical axis value
      vmax (float): the current maximum vertical axis value

    Returns:
      float, float: the new vmin and vmax
    """
    if v < vmin:
      vpad = 0.1 * (vmax - v)
      vmin = v - vpad
    if v > vmax:
      vpad = 0.1 * (v - vmin)
      vmax = v + vpad
    return vmin, vmax

  def update_plots(self, i):
    """This function is called by Matplotlib for each animation frame

    Each time called, collect FNCS messages until the next time to plot
    has been reached. Then update the plot quantities and return the
    Line2D objects that have been updated for plotting. Check for new
    data outside the plotted vertical range, which triggers a full
    re-draw of the axes. On the last frame, finalize FNCS.

    Args:
      i (int): the animation frame number
    """
#    print ('.', end='')
#    print ('frame', i, 'of', self.nsteps)
    bRedraw = False
    while self.time_granted < self.time_stop: # time in seconds
      # find the time value and index into the time (X) array
      self.time_granted = fncs.time_request(self.time_stop)
      events = fncs.get_events()
      self.root.update()
      idx = int (self.time_granted / self.yaml_delta)
      if idx <= self.idxlast:
        continue
      self.idxlast = idx
      h = float (self.time_granted / 3600.0)
      self.hrs.append (h)

      # find the newest Y values
      v0 = 0.0
      v1 = 0.0
      v2auc = 0.0
      v2lmp = 0.0
      v3 = 0.0
      for topic in events:
        value = fncs.get_value(topic)
        if topic == 'power_A':
          v1 = 3.0 * float (value.strip('+ degFkW')) / 1000.0
        elif topic == 'distribution_load':
          v3 = helpers.parse_kw (value)
          self.gld_load = v3
        elif topic == 'vpos7':
          v0 = float (value.strip('+ degFkW')) / 133000.0
        elif topic == 'clear_price':
          v2auc = float (value.strip('+ degFkW'))
        elif topic == 'LMP7':
          v2lmp = float (value.strip('+ degFkW'))
        elif topic == 'SUBSTATION7':
          v1 = float (value.strip('+ degFkW')) # already in kW

      # expand the Y axis limits if necessary, keeping a 10% padding around the range
      if v0 < self.y0min or v0 > self.y0max:
        self.y0min, self.y0max = self.expand_limits (v0, self.y0min, self.y0max)
        self.ax[0].set_ylim (self.y0min, self.y0max)
        bRedraw = True
      if v1 < self.y1min or v1 > self.y1max:
        self.y1min, self.y1max = self.expand_limits (v1, self.y1min, self.y1max)
        self.ax[1].set_ylim (self.y1min, self.y1max)
        bRedraw = True
      if v2auc > v2lmp:
        v2max = v2auc
        v2min = v2lmp
      else:
        v2max = v2lmp
        v2min = v2auc
      if v2min < self.y2min or v2max > self.y2max:
        self.y2min, self.y2max = self.expand_limits (v2min, self.y2min, self.y2max)
        self.y2min, self.y2max = self.expand_limits (v2max, self.y2min, self.y2max)
        self.ax[2].set_ylim (self.y2min, self.y2max)
        bRedraw = True
      if v3 < self.y3min or v3 > self.y3max:
        self.y3min, self.y3max = self.expand_limits (v3, self.y3min, self.y3max)
        self.ax[3].set_ylim (self.y3min, self.y3max)
        bRedraw = True

      # update the Y axis data to draw
      self.y0.append (v0) # Vpu
      self.y1.append (v1) # school kW
      self.y2auc.append (v2auc) # price
      self.y2lmp.append (v2lmp) # LMP
      self.y3fncs.append (v3) # this feeder load from FNCS (could be zero if no update)
      self.y3gld.append (self.gld_load) # most recent feeder load from FNCS
      self.ln0.set_data (self.hrs, self.y0)
      self.ln1.set_data (self.hrs, self.y1)
      self.ln2auc.set_data (self.hrs, self.y2auc)
      self.ln2lmp.set_data (self.hrs, self.y2lmp)
      self.ln3fncs.set_data (self.hrs, self.y3fncs)
      self.ln3gld.set_data (self.hrs, self.y3gld)

      if bRedraw:
#        print ('redrawing axes')
        self.fig.canvas.draw()
      if i >= (self.nsteps - 1):
        print ('finalizing FNCS')
        fncs.finalize()
        print ('FNCS active to False')
        self.bFNCSactive = False

      return self.ln0, self.ln1, self.ln2auc, self.ln2lmp, self.ln3fncs, self.ln3gld,

    print ('not finalizing FNCS')
    return self.ln0, self.ln1, self.ln2auc, self.ln2lmp, self.ln3fncs, self.ln3gld,  # in case we miss the last point

  def launch_all(self):
    """Launches the simulators, initializes FNCS and starts the animated plots
    """
    self.root.update()

    print('launching all simulators')
    self.pids = []
    for row in self.commands:
      procargs = row['args']
      if sys.platform == 'win32':
        if procargs[0] == 'python3':
          procargs[0] = 'python'  # python3 not defined on Windows
      procenv = os.environ.copy()
      for var in row['env']:
        procenv[var[0]] = var[1]
      logfd = None
      if 'log' in row:
        logfd = open (row['log'], 'w')
      proc = subprocess.Popen (procargs, env=procenv, stdout=logfd)
      self.pids.append (proc)

    print('launched', len(self.pids), 'simulators')
    self.root.update()

#    print ('want to initialize FNCS', os.environ['FNCS_CONFIG_FILE'], os.getcwd())
    fncs.initialize()
    self.bFNCSactive = True
    print ('FNCS initialized')
    self.nsteps = int (self.time_stop / self.yaml_delta)
    self.idxlast = -1
    self.time_granted = 0

    ani = animation.FuncAnimation (self.fig, self.update_plots, frames=self.nsteps,
                                   blit=True, repeat=False, interval=0)
    self.fig.canvas.draw()

def show_tesp_monitor ():
  """Creates and displays the monitor GUI
  """
  root = tk.Tk()
  root.title('Transactive Energy Simulation Platform: Solution Monitor')
  my_gui = TespMonitorGUI (root)
  root.update()
  while True:
    try:
      root.mainloop()
      break
    except UnicodeDecodeError:
      pass

